/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: netflow_info.h
* Version:  1.0
* Function Description:
* A hook for netflow info collection 
*
* History:
*   <author>     <time>     <version>    <desc>
*   maguicai  2023/05/19   1.0         Initial build
********************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/cdev.h>
#include <linux/udp.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <linux/kprobes.h>
#include <linux/inet.h>
#include <linux/net.h>
#include <linux/workqueue.h>
#include <linux/slab.h>
#include <linux/version.h>

#define TASK_COMM_LEN 16
#define MAX_PAYLOAD_BYTES 256
#define FIFO_SIZE 2048
#define RING_BUFFER_SIZE 2048
#define MAX_IP_HEADER 40
#define MAX_PACKET_NODES 100

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0))
#pragma message "Kernel version is compatible."



struct my_data
{
    struct sock * socket;
};
void process_netflow_payload(struct sk_buff *skb, int hook_num);
static int entry_handler(struct kretprobe_instance *ri, struct pt_regs *regs);
static int ret_handler(struct kretprobe_instance *ri, struct pt_regs *regs);
//static int get_inet_info_from_sock_add(struct sock *sk);
//static int get_inet_info_from_sock_del(struct sock *sk);
//static void print_tuple_for_test(const struct netflow_proc_tuple *tuple);

#define DEFINE_KRETPROBE(FUNC_NAME) \
    static struct kretprobe kretprobe_##FUNC_NAME = { \
        .entry_handler = entry_handler, \
        .handler = ret_handler, \
        .data_size = sizeof(struct my_data), \
        /* Probe up to 20 instances concurrently. */ \
        .maxactive = 20, \
        .kp.symbol_name = #FUNC_NAME, \
    }

#define REGISTER_KRETPROBES(FUNC_NAME) register_kretprobe(&kretprobe_##FUNC_NAME)
#define REGISTER_KPROBES(FUNC_NAME) register_kprobe(&kprobe_##FUNC_NAME)
#define UNREGISTER_KPROBES(FUNC_NAME) unregister_kprobe(&kprobe_##FUNC_NAME)
#define UNREGISTER_KRETPROBES(FUNC_NAME) unregister_kretprobe(&kretprobe_##FUNC_NAME)



DEFINE_KRETPROBE(tcp_connect);
DEFINE_KRETPROBE(inet_csk_accept);
DEFINE_KRETPROBE(tcp_close);
DEFINE_KRETPROBE(tcp_fin);
DEFINE_KRETPROBE(__inet_bind);
DEFINE_KRETPROBE(__inet6_bind);
DEFINE_KRETPROBE(udp_lib_unhash);
DEFINE_KRETPROBE(udp_recvmsg);
DEFINE_KRETPROBE(udpv6_recvmsg);

enum direction { INGRESS, EGRESS };


typedef enum {
        IP_VERSION_IPV4 = 0,
        IP_VERSION_IPV6
} ip_version_t;
union ip_addr {
        struct in_addr ipv4;                    // IPv4地址
        struct in6_addr ipv6;                   // IPv6地址
};
struct netflow_proc_tuple{
        union ip_addr src_ip;
        union ip_addr dst_ip;
        uint16_t src_port;
        uint16_t dst_port;
        uint8_t protocol;
        ip_version_t ip_version;
        char process_name[16];
        int  pid;
        unsigned char add;  //增加、删除
        unsigned char dir;  // 1: 出站，0：入站
        char print_ok;
        //struct work_struct save_work;
};


 struct info_work_t {
    struct work_struct save_info_work;
    struct netflow_proc_tuple data;
};
struct netflow_info_workqueue_t {
    struct workqueue_struct *info_workqueue;
    char is_valid;
    spinlock_t lock;//保护工作队列的状态

};
struct netflow_info_workqueue_t netflow_info_workqueue;
struct ring_buffer {
    struct netflow_proc_tuple buffer[RING_BUFFER_SIZE];
    unsigned int read_pos;
    unsigned int write_pos;
    struct mutex lock;
};
static struct ring_buffer rb_info;
//static struct workqueue_struct *kprobe_workqueue;
//static LIST_HEAD(packet_node_list);  //初始化队列


//static struct kmem_cache *netflow_proc_tuple_cache = NULL;






static int major_number;
static struct class *netflow_info_class;

/***
 * count 代表读取多少条 packet_data数据
 * 返回值：读取了多少条数据
*/
static ssize_t netflow_chardev_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
    //printk(KERN_INFO " netflow read from ko  count=%d...\n",count);
    ssize_t total = 0 ;
    int available;
    if (mutex_lock_interruptible(&rb_info.lock)) {
    return -ERESTARTSYS;
    }
    //需要检查 这个count是否有效，避免访问非法内存
    available = (rb_info.write_pos >= rb_info.read_pos) ? 
    (rb_info.write_pos - rb_info.read_pos) : (RING_BUFFER_SIZE - rb_info.read_pos + rb_info.write_pos);

    
    // 如果请求的数目比可用的数据多时，取可用数
    if (count > available) {
        count = available;

        /***
        mutex_unlock(&rb.lock);
        return -EINVAL;
        **/
    }
    //printk(KERN_INFO " netflow read from ko  count=%d...\n",count);
    


    while (count > 0 && rb_info.read_pos != rb_info.write_pos) {
        //printk(KERN_INFO " netflow read from ko Current read_pos=%d write_pos = %d total=%d,count=%d...\n",rb_info.read_pos, rb_info.write_pos,total,count);
        // buf是char *指针，buffer是 struct packet_data数组,注意这个转换
        if(copy_to_user(buf + total * sizeof(struct netflow_proc_tuple), rb_info.buffer + rb_info.read_pos, sizeof(struct netflow_proc_tuple))){
		    return -EFAULT;
	    }	
        //buf[total * sizeof(struct packet_data)] = rb.buffer[rb.read_pos];
        rb_info.read_pos = (rb_info.read_pos + 1) % RING_BUFFER_SIZE;
        total++;
        count--;
        //printk(KERN_INFO " netflow read from ko Next read_pos=%d write_pos = %d total=%d,count=%d...\n",rb.read_pos, rb.write_pos,total,count);
    }
    mutex_unlock(&rb_info.lock);
    return total;
}
static const struct file_operations netflow_info_fops = {
    .owner = THIS_MODULE,
    .read = netflow_chardev_read,
};



static int __init chardev_init(void)
{
    major_number = register_chrdev(0, "netflow_info", &netflow_info_fops);
    if (major_number < 0)
    {
        printk(KERN_ERR "register_chrdev failed: %d\n", major_number);
        return major_number;
    }

    netflow_info_class = class_create(THIS_MODULE, "netflow_info_chardev_class");
    if (IS_ERR(netflow_info_class))
    {
        unregister_chrdev(major_number, "netflow_info");
        return PTR_ERR(netflow_info_class);
    }

    device_create(netflow_info_class, NULL, MKDEV(major_number, 0), NULL, "netflow_info");
    printk(KERN_INFO "Kprobe CharDEV created ...\n");

    return 0;
}

static void __exit chardev_exit(void)
{
    device_destroy(netflow_info_class, MKDEV(major_number, 0));
    class_destroy(netflow_info_class);
    unregister_chrdev(major_number, "netflow_info");
}




static void print_tuple_for_test(const struct netflow_proc_tuple *tuple, int func_index){

    char src_ip[INET6_ADDRSTRLEN] = {0};
    char dst_ip[INET6_ADDRSTRLEN] = {0};

    if (tuple->ip_version == IP_VERSION_IPV4) {
        snprintf(src_ip, sizeof(src_ip), "%pI4", &tuple->src_ip.ipv4);
        snprintf(dst_ip, sizeof(dst_ip), "%pI4", &tuple->dst_ip.ipv4);
    } else {
        snprintf(src_ip, sizeof(src_ip), "%pI6", &tuple->src_ip.ipv6);
        snprintf(dst_ip, sizeof(dst_ip), "%pI6", &tuple->dst_ip.ipv6);
    }

    pr_info("Src IP: %s, Dst IP: %s\n", src_ip, dst_ip);
    pr_info("Src port: %u/%u, Dst port: %u/%u\n", ntohs(tuple->src_port),tuple->src_port, ntohs(tuple->dst_port),tuple->dst_port);
    pr_info("Protocol: %s\n", tuple->protocol == 17 ? "UDP" : "TCP");
    pr_info("IP version: %s\n", tuple->ip_version == IP_VERSION_IPV4 ? "IPv4" : "IPv6");
    pr_info("Process name: %s, PID: %d\n", tuple->process_name, tuple->pid);
    pr_info("Add: %s\n", tuple->add ? "ADD" : "DEL");
    pr_info("Direction: %s\n", tuple->dir ? "Outbound" : "Inbound");
    pr_info("Function: %d\n", func_index);
}
static int save_tuple_info(struct netflow_proc_tuple *tuple){

    if(tuple == NULL){
        printk(KERN_ERR " save_tuple_info error netflow_proc_tuple is NULL !! \n");
        return -1;
    }
    //print_tuple_for_test(tuple,5);
    mutex_lock(&rb_info.lock);
    //printk(KERN_INFO " Saving tuple info Before  :  write_pos=%d read_pos=%d \n",rb_info.write_pos,rb_info.read_pos);
    memcpy(&rb_info.buffer[rb_info.write_pos],tuple,sizeof(struct netflow_proc_tuple));
    rb_info.write_pos = (rb_info.write_pos + 1) % RING_BUFFER_SIZE;
    //检查 是否覆盖了读指针
    if(rb_info.write_pos == rb_info.read_pos){
        rb_info.read_pos = (rb_info.read_pos + 1) % RING_BUFFER_SIZE;
    }
    //printk(KERN_INFO " Saving tuple info End :  write_pos=%d read_pos=%d \n",rb_info.write_pos,rb_info.read_pos);

    //print_tuple_for_test(&rb_info.buffer[rb_info.write_pos-1],6);

    mutex_unlock(&rb_info.lock);

    return 0;
}

void netflow_wq_save_info(struct work_struct *work){
    struct info_work_t *info_work = (struct info_work_t *)work;
    // save data to buffer 
    struct netflow_proc_tuple *tuple = &info_work->data;
    save_tuple_info(tuple);

    //printk(KERN_INFO "Work queue: data saved! \n" );

    kfree ((void *)work);
}
static inline int get_tuple_pid_info_tcp(struct pt_regs *ctx, struct sock *sk, unsigned char add){
	struct inet_sock *inet = (struct inet_sock *)sk;
	//获取端口
	//u16 dport, sport;
	u16 src_port,dst_port;
	u16 family;
    struct info_work_t *save_work ;
	struct netflow_proc_tuple data = {0};
	src_port=inet->inet_sport;
	dst_port = inet->sk.__sk_common.skc_dport;

	//src_port = bpf_ntohs(sport);
	//dst_port = bpf_ntohs(dport);
	//bpf_printk("sport=%d dport=%d\n",src_port,dst_port);
	//获取协议族
	family = sk->__sk_common.skc_family;

    save_work = (struct info_work_t*)kmalloc(sizeof(struct info_work_t), GFP_ATOMIC);
    if(save_work){
        INIT_WORK((struct work_struct *)save_work, netflow_wq_save_info);
    }

	//准备结构体数据
    //INIT_WORK(&data.save_work, save_info);
	data.add = add;
    data.pid = task_pid_nr(current);
    strncpy(data.process_name, current->comm, TASK_COMM_LEN);
    data.process_name[TASK_COMM_LEN - 1] = '\0';
	data.src_port = src_port;
	data.dst_port = dst_port;
	if(family == 2){
		data.ip_version = IP_VERSION_IPV4;
		data.src_ip.ipv4.s_addr = sk->sk_rcv_saddr;
        data.dst_ip.ipv4.s_addr = sk->__sk_common.skc_daddr;	
	}else if (family ==  10){
		data.ip_version = IP_VERSION_IPV6;
        data.src_ip.ipv6 = sk->__sk_common.skc_v6_rcv_saddr;
   		data.dst_ip.ipv6 = sk->__sk_common.skc_v6_daddr;
	}
	data.protocol = IPPROTO_TCP;
   	//bpf_core_read(&data.protocol, sizeof(data.protocol), &(sk->sk_protocol));
    
    //for test
    //print_tuple_for_test(&data,0);
    //save_tuple_info(&data);
    // 交给工作队列去保存数据
    //save_work->data = data;
    memcpy(&save_work->data,&data,sizeof(struct netflow_proc_tuple));
    spin_lock(&netflow_info_workqueue.lock);
    if (!netflow_info_workqueue.is_valid){
        spin_unlock(&netflow_info_workqueue.lock);
        return 0;
    }
    spin_unlock(&netflow_info_workqueue.lock);
    queue_work(netflow_info_workqueue.info_workqueue,(struct work_struct *)save_work);
	return 0;

}
static inline int get_tuple_pid_info_tcp_accept(struct pt_regs *ctx, struct sock *sk, unsigned char add){
	struct inet_sock *inet = (struct inet_sock *)sk;
	//获取端口
	//u16 dport, sport;
	u16 src_port,dst_port;
	u16 family;
     struct info_work_t *save_work ;
	struct netflow_proc_tuple data = {0};
	dst_port = inet->inet_sport;
	src_port = inet->sk.__sk_common.skc_dport;

	//src_port = bpf_ntohs(sport);
	//dst_port = bpf_ntohs(dport);
	//bpf_printk("sport=%d dport=%d\n",src_port,dst_port);
	//获取协议族
	family = sk->__sk_common.skc_family;


    save_work = (struct info_work_t*)kmalloc(sizeof(struct info_work_t), GFP_ATOMIC);
    if(save_work){
        INIT_WORK((struct work_struct *)save_work, netflow_wq_save_info);
    }
	//准备结构体数据
    //INIT_WORK(&data.save_work, save_info);
	data.add = add;
    data.pid = task_pid_nr(current);
    strncpy(data.process_name, current->comm, TASK_COMM_LEN);
    data.process_name[TASK_COMM_LEN - 1] = '\0';
	data.src_port = src_port;
	data.dst_port = dst_port;
	if(family == 2){
		data.ip_version = IP_VERSION_IPV4;
		data.dst_ip.ipv4.s_addr = sk->__sk_common.skc_rcv_saddr;
        data.src_ip.ipv4.s_addr = sk->__sk_common.skc_daddr;	
	}else if (family ==  10){
		data.ip_version = IP_VERSION_IPV6;
        data.dst_ip.ipv6 = sk->__sk_common.skc_v6_rcv_saddr;
   		data.src_ip.ipv6 = sk->__sk_common.skc_v6_daddr;
	}
	data.protocol = IPPROTO_TCP;
   	//bpf_core_read(&data.protocol, sizeof(data.protocol), &(sk->sk_protocol));
    //for test
    //print_tuple_for_test(&data,1);
    //save_tuple_info(&data);
    memcpy(&save_work->data,&data,sizeof(struct netflow_proc_tuple));
    spin_lock(&netflow_info_workqueue.lock);
    if (!netflow_info_workqueue.is_valid){
        spin_unlock(&netflow_info_workqueue.lock);
        return 0;
    }
    spin_unlock(&netflow_info_workqueue.lock);
    queue_work(netflow_info_workqueue.info_workqueue,(struct work_struct *)save_work);
    
    // 交给工作队列去保存数据
    //queue_work(kprobe_workqueue,&data.save_work);
	return 0;

}
static inline int get_tuple_pid_info_udp_unhash(struct pt_regs *ctx, struct sock *sk, unsigned char add){
	struct inet_sock *inet = (struct inet_sock *)sk;
//获取端口
	//u16 dport, sport;
	u16 src_port,dst_port;
	u16 family;
	struct netflow_proc_tuple data = {0};
    struct info_work_t *save_work ;
	src_port = inet->inet_sport;
	dst_port = inet->sk.__sk_common.skc_dport;

	//src_port = bpf_ntohs(sport);
	//dst_port = bpf_ntohs(dport);
    	//bpf_printk("sport=%d dport=%d\n",src_port,dst_port);
	//获取协议族
	family = sk->__sk_common.skc_family;

    save_work = (struct info_work_t*)kmalloc(sizeof(struct info_work_t), GFP_ATOMIC);
    if(save_work){
        INIT_WORK((struct work_struct *)save_work, netflow_wq_save_info);
    }

	//准备结构体数据
    //INIT_WORK(&(data.save_work), save_info);
	data.add = add;
    data.pid = task_pid_nr(current);
    strncpy(data.process_name, current->comm, TASK_COMM_LEN);
    data.process_name[TASK_COMM_LEN - 1] = '\0';
	data.src_port = src_port;
	data.dst_port = dst_port;
	if(family == 2){
		data.ip_version = IP_VERSION_IPV4;
		data.src_ip.ipv4.s_addr = sk->__sk_common.skc_rcv_saddr;
        data.dst_ip.ipv4.s_addr = sk->__sk_common.skc_daddr;	
	}else if (family ==  10){
		data.ip_version = IP_VERSION_IPV6;
        data.src_ip.ipv6 = sk->__sk_common.skc_v6_rcv_saddr;
   		data.dst_ip.ipv6 = sk->__sk_common.skc_v6_daddr;
	}
	data.protocol = IPPROTO_UDP;
    //for test
    //print_tuple_for_test(&data,2);
    //save_tuple_info(&data);
    
    
    // 交给工作队列去保存数据
    memcpy(&save_work->data,&data,sizeof(struct netflow_proc_tuple));
    spin_lock(&netflow_info_workqueue.lock);
    if (!netflow_info_workqueue.is_valid){
        spin_unlock(&netflow_info_workqueue.lock);
        return 0;
    }
    spin_unlock(&netflow_info_workqueue.lock);
    queue_work(netflow_info_workqueue.info_workqueue,(struct work_struct *)save_work);
    //queue_work(kprobe_workqueue,&data.save_work);
	return 0;

}
static inline int get_tuple_pid_info_udp_recvmsg(struct pt_regs *ctx, struct sock *sk,void *msg_name, unsigned char add){
	struct inet_sock *inet = (struct inet_sock *)sk;
	// 获取进程信息
	struct sockaddr_in *addr;
	u32 src_ip ; // 源IP地址
    struct info_work_t *save_work ;
	struct netflow_proc_tuple data = {0};
	//u32 dst_ip ;   // 目标IP地址
	u16 src_port ;  // 源端口
	u16 dst_port  ; // 目标端口
	u16 family;
	addr = (struct sockaddr_in *)msg_name;
	src_ip = sk->__sk_common.skc_rcv_saddr;
    dst_port = inet->inet_sport;
	src_port = inet->sk.__sk_common.skc_dport;
	family = sk->__sk_common.skc_family;

    save_work = (struct info_work_t*)kmalloc(sizeof(struct info_work_t), GFP_ATOMIC);
    if(save_work){
        INIT_WORK((struct work_struct *)save_work, netflow_wq_save_info);
    }
	//准备结构体数据
    //INIT_WORK(&(data.save_work), save_info);
	data.add = add;
	data.dir = 1;
    data.pid = task_pid_nr(current);
    strncpy(data.process_name, current->comm, TASK_COMM_LEN);
    data.process_name[TASK_COMM_LEN - 1] = '\0';
	data.src_port = src_port;
	data.dst_port = dst_port;
	if(family == 2){
		data.ip_version = IP_VERSION_IPV4;
		data.dst_ip.ipv4.s_addr = sk->__sk_common.skc_rcv_saddr;
        data.src_ip.ipv4.s_addr = sk->__sk_common.skc_daddr;	
	}else if (family ==  10){
		data.ip_version = IP_VERSION_IPV6;
        data.dst_ip.ipv6 = sk->__sk_common.skc_v6_rcv_saddr;
   		data.src_ip.ipv6 = sk->__sk_common.skc_v6_daddr;
	}
	data.protocol = IPPROTO_UDP;
    //for test
    //print_tuple_for_test(&data,3);
    
    // 交给工作队列去保存数据
    memcpy(&save_work->data,&data,sizeof(struct netflow_proc_tuple));
    spin_lock(&netflow_info_workqueue.lock);
    if (!netflow_info_workqueue.is_valid){
        spin_unlock(&netflow_info_workqueue.lock);
        return 0;
    }
    spin_unlock(&netflow_info_workqueue.lock);
    queue_work(netflow_info_workqueue.info_workqueue,(struct work_struct *)save_work);
    //printk(KERN_INFO "put save_work to queue!!!\n");
    //save_tuple_info(&data);
    //queue_work(kprobe_workqueue,&data.save_work);
	return 0;

}

static int entry_handler(struct kretprobe_instance *ri, struct pt_regs *regs)
{
    //int size;
    struct sock *sk;
    struct kretprobe *rp = ri->rp; 

    if(rp ==  &kretprobe_tcp_connect ){

#ifdef CONFIG_X86
        sk = (struct sock *)regs->di;
#endif
#ifdef CONFIG_ARM64
        sk = (struct sock *)regs->regs[0];
#endif
#ifdef CONFIG_LOONGARCH
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_MIPS
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_SW64
	#ifdef CONFIG_ACPI_GENERIC_GSI
        	sk = (struct sock *)regs->regs[16];
	#else
        	sk = (struct sock *)regs->r16;
	#endif
#endif
    if(sk){
        get_tuple_pid_info_tcp(regs,sk,1);
        return 0;
        }
    }
    else if(rp == &kretprobe_udp_lib_unhash){

#ifdef CONFIG_X86
        sk = (struct sock *)regs->di;
#endif
#ifdef CONFIG_ARM64
        sk = (struct sock *)regs->regs[0];
#endif
#ifdef CONFIG_LOONGARCH
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_MIPS
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_SW64
	#ifdef CONFIG_ACPI_GENERIC_GSI
                sk = (struct sock *)regs->regs[16];
        #else
                sk = (struct sock *)regs->r16;
        #endif
#endif
        if(sk){
            get_tuple_pid_info_udp_unhash(regs,sk,0);
            return 0;
        }
    }
    else if(rp == &kretprobe_inet_csk_accept  ){
        // 第一个参数同样是struct sock * 类型, 新增端口的情况
#ifdef CONFIG_X86
        sk = (struct sock *)regs->di;
#endif
#ifdef CONFIG_ARM64
        sk = (struct sock *)regs->regs[0];
#endif
#ifdef CONFIG_LOONGARCH
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_MIPS
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_SW64
	#ifdef CONFIG_ACPI_GENERIC_GSI
                sk = (struct sock *)regs->regs[16];
        #else
                sk = (struct sock *)regs->r16;
        #endif
#endif
        if(sk){
            get_tuple_pid_info_tcp_accept(regs, sk, 1);
            return 0;
        }
    }else if(rp  == &kretprobe_tcp_close || rp == &kretprobe_tcp_fin){
        // 第一个参数 是struct sock * 类型，释放端口的情况
#ifdef CONFIG_X86
        sk = (struct sock *)regs->di;
#endif
#ifdef CONFIG_ARM64
        sk = (struct sock *)regs->regs[0];
#endif
#ifdef CONFIG_LOONGARCH
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_MIPS
        sk = (struct sock *)regs->regs[4];
#endif
#ifdef CONFIG_SW64
	#ifdef CONFIG_ACPI_GENERIC_GSI
                sk = (struct sock *)regs->regs[16];
        #else
                sk = (struct sock *)regs->r16;
        #endif
#endif
        if(sk){
            get_tuple_pid_info_tcp(regs,sk,0);
        }

    }
    else if(rp  == &kretprobe_udp_recvmsg || rp  == &kretprobe_udpv6_recvmsg){
        // 第一个参数 是struct sock * 类型，释放端口的情况
        struct msghdr *msg ;
#ifdef CONFIG_X86
        sk = (struct sock *)regs->di;
        msg = (struct msghdr *)regs->si;
#endif
#ifdef CONFIG_ARM64
        sk = (struct sock *)regs->regs[0];
        msg = (struct msghdr *)regs->regs[1];
#endif
#ifdef CONFIG_LOONGARCH
        sk = (struct sock *)regs->regs[4];
        msg = (struct msghdr *)regs->regs[5];
#endif
#ifdef CONFIG_MIPS
        sk = (struct sock *)regs->regs[4];
        msg = (struct msghdr *)regs->regs[5];
#endif
#ifdef CONFIG_SW64
	#ifdef CONFIG_ACPI_GENERIC_GSI
                sk = (struct sock *)regs->regs[16];
        	msg = (struct msghdr *)regs->regs[17];
        #else
                sk = (struct sock *)regs->r16;
        	msg = (struct msghdr *)regs->r17;
        #endif
#endif
        if(sk){
            get_tuple_pid_info_udp_recvmsg(regs,sk,msg,1);
        }

    }
    /**
    else if(rp == &kretprobe___inet_bind ){
        // 使用第二个参数 struct sockaddr *uaddr ,新增端口
#ifdef CONFIG_X86
        struct sockaddr_in *addr = (struct sockaddr_in *)regs->si;
#endif
#ifdef CONFIG_ARM64
        struct sockaddr_in *addr = (struct sockaddr_in *)regs->regs[1];
#endif
#ifdef CONFIG_LOONGARCH
        struct sockaddr_in *addr = (struct sockaddr_in *)regs->regs[5];
#endif
        if(addr && addr->sin_port){
            size = snprintf(msg, MESSAGE_SIZE, "%d:%d:%hu",PORT_ADD_IPv4 ,current->pid , htons(addr->sin_port));
            size = snprintf(message, MESSAGE_SIZE, "%d:%s\n",size,msg);
            message_in(message,size);
        }
    }else if(rp == &kretprobe___inet6_bind){
        // 使用第二个参数 struct sockaddr *uaddr ,新增端口
#ifdef CONFIG_X86
        struct sockaddr_in6 *addr = (struct sockaddr_in6 *)regs->si;
#endif
#ifdef CONFIG_ARM64
        struct sockaddr_in6 *addr = (struct sockaddr_in6 *)regs->regs[1];
#endif
#ifdef CONFIG_LOONGARCH
        struct sockaddr_in6 *addr = (struct sockaddr_in6 *)regs->regs[5];
#endif
      if(sk){
        get_tuple_pid_info_tcp(regs,sk,1);
        return 0;
        }
        

    }
    **/
    return 0;
}

static int ret_handler(struct kretprobe_instance *ri, struct pt_regs *regs)
{
    return 0;
}


int kretprobe_init(void)
{

    // 初步怀疑，在udp_lib_get_port上使用kretprobe时会引起软锁死机。
    if (REGISTER_KRETPROBES(tcp_connect) < 0 ||
        REGISTER_KRETPROBES(inet_csk_accept) < 0 ||
        REGISTER_KRETPROBES(tcp_fin) < 0 ||
        REGISTER_KRETPROBES(tcp_close) < 0 ||
        REGISTER_KRETPROBES(__inet_bind) < 0 ||
        REGISTER_KRETPROBES(__inet6_bind) < 0 ||
        REGISTER_KRETPROBES(udp_recvmsg) < 0 ||
        REGISTER_KRETPROBES(udpv6_recvmsg) < 0 ||
        REGISTER_KRETPROBES(udp_lib_unhash) < 0 
        ){
        pr_err("register_kretprobe failed");
     }
    return 0;
}
void kretprobe_exit(void)
{
    UNREGISTER_KRETPROBES(tcp_connect);
    UNREGISTER_KRETPROBES(inet_csk_accept);
    UNREGISTER_KRETPROBES(tcp_close);
    UNREGISTER_KRETPROBES(tcp_fin);
    UNREGISTER_KRETPROBES(__inet_bind);
    UNREGISTER_KRETPROBES(__inet6_bind);
    UNREGISTER_KRETPROBES(udp_lib_unhash);
    UNREGISTER_KRETPROBES(udp_recvmsg);
    UNREGISTER_KRETPROBES(udpv6_recvmsg);
}
static int __init nf_info_init(void)
{
    int ret;

    // 初始化ring_buffer 
    memset(rb_info.buffer, 0, sizeof(rb_info.buffer));
    rb_info.read_pos = 0;
    rb_info.write_pos = 0;
    mutex_init(&rb_info.lock);


  
    spin_lock_init(&netflow_info_workqueue.lock);
   netflow_info_workqueue.info_workqueue = create_singlethread_workqueue("save_info_workqueue");
   netflow_info_workqueue.is_valid = 1;
   if (!netflow_info_workqueue.info_workqueue){
    return -ENOMEM;
   }


    printk(KERN_INFO "Kprobe Netfilter hooks registered\n");
    ret = chardev_init();
    printk(KERN_INFO "Kprobe Netfilter CharDev init..\n");
    kretprobe_init();
    printk(KERN_INFO "Kprobe Netflow Kretprobe init..\n");
    return 0;
}

static void __exit nf_info_exit(void)
{


    if (netflow_info_workqueue.info_workqueue){
        spin_lock(&netflow_info_workqueue.lock);
        netflow_info_workqueue.is_valid = 0;
        spin_unlock(&netflow_info_workqueue.lock);
        flush_workqueue(netflow_info_workqueue.info_workqueue);
        destroy_workqueue(netflow_info_workqueue.info_workqueue);
        netflow_info_workqueue.info_workqueue = NULL;
    }
     chardev_exit();
    printk(KERN_INFO "Netfilter CharDEV unregistered\n");
    kretprobe_exit();
    printk(KERN_INFO "Netflow kretprobe unregistered\n");
}


#else
#warning "Kernel version is not comptible"

static int __init nf_info_init(void)
{

    printk(KERN_INFO "Netflow ktprobe register is not comptible with Current Kernel Version !\n");
    return 0;
}

static void __exit nf_info_exit(void)
{

    printk(KERN_INFO "Netflow ktprobe unregister is not comptible with Current Kernel Version !\n");
}

#endif



module_init(nf_info_init);
module_exit(nf_info_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kylin");
MODULE_DESCRIPTION("Netfilter hooks for Netflow payload");
