/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: hashtab.h
* Version:  1.0
* Function Description:
*	A hash table (hashtab) maintains associations between
*	key values and datum values.  The type of the key values
*	functions for hash computation and key comparison are
*	provided by the creator of the table.
*
* History:
*   <author>     <time>     <version>    <desc>
*   wangshengye  2023/05/19   1.0         Initial build
********************************************************************************/

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <net/ipv6.h>
#include "common.h"
#include "hashtab.h"

static struct kmem_cache *hashtab_node_cachep;
struct hashtab *policy_hashtab = NULL;

#define hashtab_node_free(node)                 \
	kfree(node->key);                           \
	kfree(node->datum);                         \
	kmem_cache_free(hashtab_node_cachep, node); \
	node = NULL;

#define lru_node_move(node, head)     \
	node->next = head;                \
	node->prev = NULL;                \
	node->capacity = head->capacity;  \
	head->prev = node;                \
	head->tail = NULL;

struct hashtab *hashtab_create(u32 (*hash_value)(struct hashtab *h, const void *key),
			       int (*keycmp)(struct hashtab *h, const void *key1, const void *key2),
			       u32 size)
{
	struct hashtab *p;
	u32 i;

	if ( !hash_value || !keycmp)
		return NULL;

	p = kzalloc(sizeof(*p), GFP_KERNEL);
	if (!p)
		return p;

	p->size = size;
	p->nel = 0;
	p->hash_value = hash_value;
	p->keycmp = keycmp;
	p->htable = kmalloc_array(size, sizeof(*p->htable), GFP_KERNEL);
	if (!p->htable) {
		kfree(p);
		return NULL;
	}

	for (i = 0; i < size; i++)
		p->htable[i] = NULL;
	
	return p;
}

u32 hash_value(struct hashtab *h, const void *key) {
    u32 val = 0;
    struct pkt_addr_info *pkt = (struct pkt_addr_info*)key;

    if (pkt->ip_version == IP_ADDR_VERSION_IPV4) {
        val = pkt->saddr.ipv4.s_addr + pkt->daddr.ipv4.s_addr + pkt->sport + pkt->dport + pkt->dir + pkt->ip_version + pkt->protocol;
    } else if (pkt->ip_version == IP_ADDR_VERSION_IPV6) {
        val = ipv6_addr_hash(&pkt->saddr.ipv6) + ipv6_addr_hash(&pkt->daddr.ipv6) + pkt->sport + pkt->dport + pkt->dir + pkt->ip_version + pkt->protocol;
    }
    return val % h->size;
}

int key_cmp(struct hashtab *h, const void *key1, const void *key2)
{	
	struct pkt_addr_info *pkt1 = (struct pkt_addr_info *)key1;
	struct pkt_addr_info *pkt2 = (struct pkt_addr_info *)key2;

	if (!memcmp(&pkt1->sport, &pkt2->sport, sizeof(pkt2->sport)) &&
		!memcmp(&pkt1->dport, &pkt2->dport, sizeof(pkt2->sport)) &&
		pkt1->dir ==  pkt2->dir &&
		pkt1->ip_version == pkt2->ip_version &&
		pkt1->protocol == pkt2->protocol) {
		
		if (pkt1->ip_version == IP_ADDR_VERSION_IPV4) {
			if (!memcmp(&pkt1->saddr.ipv4, &pkt2->saddr.ipv4, sizeof(struct in_addr)) &&
				!memcmp(&pkt1->daddr.ipv4, &pkt2->daddr.ipv4, sizeof(struct in_addr)))
				return 0;
		} else if (pkt1->ip_version == IP_ADDR_VERSION_IPV6) {
			if (!memcmp(&pkt1->saddr.ipv6, &pkt2->saddr.ipv6, sizeof(struct in6_addr)) &&
				!memcmp(&pkt1->daddr.ipv6, &pkt2->daddr.ipv6, sizeof(struct in6_addr))) 
				return 0;
		}
	}
    /*memcmp比较结构体需要注意字节填充问题，填充的内容可能是随机值*/
	//return memcmp(key1, key2, sizeof(struct pkt_addr_info)) == 0 ? 0 : 1;
	return 1;
}

int hashtab_insert(struct hashtab *h, void *key, unsigned int key_size, 
					void *datum, unsigned int val_size)
{
	u32 hvalue;
	u32 capacity;
	struct hashtab_node *cur, *newnode, *head;

	if (!h || !key | !datum || h->nel == HASHTAB_MAX_NODES)
		return -EINVAL;

	hvalue = h->hash_value(h, key);
	head = h->htable[hvalue];
	cur = head;

    /* creates a new node as the head when the list is empty*/
	if (!head) {
		head = kmem_cache_zalloc(hashtab_node_cachep, GFP_KERNEL);
		if (!head)
			return -ENOMEM;

		memset(head, 0, sizeof(struct hashtab_node));
		head->capacity++;

		head->key = kmalloc(key_size, GFP_KERNEL);
		head->datum = kmalloc(val_size, GFP_KERNEL);
		memcpy(head->key, key, key_size);
		memcpy(head->datum, datum, val_size);

		head->tail = head;
		h->htable[hvalue] = head;

		return 0;
	}

	while (cur) {
		if (h->keycmp(h, key, cur->key) == 0) {
			if (!memcmp(cur->datum, datum, val_size)) {
				return -EEXIST;
			} else {
				memcpy(cur->datum, datum, val_size);
				lru_node_move(cur, head);
				return 0;
			}
		}
		cur = cur->next;
	}

	capacity = head->capacity;
	
	if (capacity >= HASH_SLOT_CAPACITY_MAX) {
		struct hashtab_node *tail = head->tail;
		head->tail = tail->prev;
		tail->prev->next = NULL;
		hashtab_node_free(tail);
		head->capacity--;
		h->nel--;
	}

	newnode = kmem_cache_zalloc(hashtab_node_cachep, GFP_KERNEL);
	if (!newnode)
		return -ENOMEM;
	
	memset(newnode, 0, sizeof(struct hashtab_node));

	newnode->key = kmalloc(key_size, GFP_KERNEL);
	newnode->datum = kmalloc(val_size, GFP_KERNEL);
	
	memcpy(newnode->key, key, key_size);
	memcpy(newnode->datum, datum, val_size);

	head->prev = newnode;
	newnode->next = head;
	newnode->capacity = capacity + 1;
    newnode->tail = head->tail;
	head->tail = NULL;
	h->htable[hvalue] = newnode;
	h->nel++;
	
	return 0;
}

int hashtab_remove(struct hashtab *h, const void *key)
{
	u32 hvalue;
	struct hashtab_node *cur, *head;

	if (!h || !key)
		return -EINVAL;

	hvalue = h->hash_value(h, key);
	head = h->htable[hvalue];
	cur = head;

	while (cur) {
		if (h->keycmp(h, key, cur->key) == 0) {
			if (head != cur && cur != head->tail) {
				cur->prev->next = cur->next;
				cur->next->prev = cur->prev;
			} else if (cur == head->tail && head->capacity > 1) {
				cur->prev->next = NULL;
				head->tail = cur->prev;
			} else {
				if (cur->next) {
					cur->next->tail = cur->tail;
					cur->next->capacity = cur->capacity;
					cur->next->prev = NULL;
					h->htable[hvalue] = cur->next;
				} else {
					h->htable[hvalue] = NULL;
				}
			}

			hashtab_node_free(cur);

			if (h->htable[hvalue]) {
				h->htable[hvalue]->capacity--;
				h->nel--;
			}

			return 0;
		}

		cur = cur->next;
	}

	return -ENOENT;
}

void *hashtab_search(struct hashtab *h, const void *key)
{
	u32 hvalue;
	struct hashtab_node *cur, *head;

	if (!h || !key)
		return NULL;

	hvalue = h->hash_value(h, key);
	head = h->htable[hvalue];
	cur = head;

	while (cur) {
		if (h->keycmp(h, key, cur->key) == 0) {
			if (cur != head && cur != head->tail) {
				cur->prev->next = cur->next;
				cur->next->prev = cur->prev;
				cur->tail = head->tail;

				lru_node_move(cur, head);
			} else if (cur == head->tail && head->capacity > 1) {
				cur->prev->next = NULL;
				cur->tail = cur->prev;

				lru_node_move(cur, head);
			}

			h->htable[hvalue] = cur;
			return cur->datum;
		}

		cur = cur->next;
	}

	return NULL;
}

void hashtab_cleanup(struct hashtab *h)
{
	u32 i;
	struct hashtab_node *cur, *temp;

	if (!h)
		return;

	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		while (cur) {
			temp = cur;
			cur = cur->next;

			kfree(temp->key);
			kfree(temp->datum);
			kmem_cache_free(hashtab_node_cachep, temp);
		}
		h->htable[i] = NULL;
	}
}

void hashtab_destroy(struct hashtab *h)
{
	if (!h)
		return;

	hashtab_cleanup(h);

	kfree(h->htable);
	h->htable = NULL;

	kfree(h);
}

int hashtab_map(struct hashtab *h,
		int (*apply)(void *k, void *d, void *args),
		void *args)
{
	u32 i;
	int ret;
	struct hashtab_node *cur;

	if (!h)
		return 0;

	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		while (cur) {
			ret = apply(cur->key, cur->datum, args);
			if (ret)
				return ret;
			cur = cur->next;
		}
	}
	return 0;
}

void hashtab_stat(struct hashtab *h, struct hashtab_info *info)
{
	u32 i, chain_len, slots_used, max_chain_len;
	struct hashtab_node *cur;

	slots_used = 0;
	max_chain_len = 0;
	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		if (cur) {
			slots_used++;
			chain_len = 0;
			while (cur) {
				chain_len++;
				cur = cur->next;
			}

			if (chain_len > max_chain_len)
				max_chain_len = chain_len;
		}
	}

	info->slots_used = slots_used;
	info->max_chain_len = max_chain_len;
}

void __init hashtab_cache_init(void)
{
		hashtab_node_cachep = kmem_cache_create("hashtab_node",
			sizeof(struct hashtab_node),
			0, SLAB_PANIC, NULL);
}

