/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: hashtab.h
* Version:  1.0
* Function Description:
*	A hash table (hashtab) maintains associations between
*	key values and datum values.  The type of the key values
*	functions for hash computation and key comparison are
*	provided by the creator of the table.
*
* History:
*   <author>     <time>     <version>    <desc>
*   wangshengye  2023/05/19   1.0         Initial build
********************************************************************************/
#ifndef _HASHTAB_H_
#define _HASHTAB_H_

#define HASHTB_SLOT_MAX          0x1000
#define HASH_SLOT_CAPACITY_MAX   0x100
#define HASHTAB_MAX_NODES	     (HASHTB_SLOT_MAX * HASH_SLOT_CAPACITY_MAX)

extern struct hashtab *policy_hashtab;

/*
 * hash table node entry
 * 
 */

struct hashtab_node {
    int  capacity;
	void *key;
	void *datum;
	struct hashtab_node *prev;
	struct hashtab_node *next;
	struct hashtab_node *tail;
};

struct hashtab {
	u32 size;			/* number of slots in hash table */
	u32 nel;			/* number of elements in hash table */	
	struct hashtab_node **htable;	/* hash table */
	u32 (*hash_value)(struct hashtab *h, const void *key);  /* hash function */
	int (*keycmp)(struct hashtab *h, const void *key1, const void *key2);   /* key comparison function */
};

struct hashtab_info {
	u32 slots_used;
	u32 max_chain_len;
};

/*
 * Creates a new hash table with the specified characteristics.
 *
 * Returns NULL if insufficent space is available or
 * the new hash table otherwise.
 */
struct hashtab *hashtab_create(u32 (*hash_value)(struct hashtab *h, const void *key),
			       int (*keycmp)(struct hashtab *h, const void *key1, const void *key2),
			       u32 size);

u32 hash_value(struct hashtab *h, const void *key);

int key_cmp(struct hashtab *h, const void *key1, const void *key2);

/*
 * Inserts the specified (key, datum) pair into the specified hash table.
 *
 * Returns -ENOMEM on memory allocation error,
 * -EEXIST if there is already an entry with the same key,
 * -EINVAL for general errors or
  0 otherwise.
 */
int hashtab_insert(struct hashtab *h, void *k, unsigned int key_size,
					void *d, unsigned int val_size);

/*
 * Remove the specified (key, datum) pair into the specified hash table.
 *
 * Returns -ENOMEM on memory allocation error,
 * -EEXIST if there is already an entry with the same key,
 * -EINVAL for general errors or
  0 otherwise.
 */

int hashtab_remove(struct hashtab *h, const void *key);

/*
 * Searches for the entry with the specified key in the hash table.
 *
 * Returns NULL if no entry has the specified key or
 * the datum of the entry otherwise.
 */
void *hashtab_search(struct hashtab *h, const void *k);

void hashtab_cleanup(struct hashtab *h);

/*
 * Destroys the specified hash table.
 */
void hashtab_destroy(struct hashtab *h);

/*
 * Applies the specified apply function to (key,datum,args)
 * for each entry in the specified hash table.
 *
 * The order in which the function is applied to the entries
 * is dependent upon the internal structure of the hash table.
 *
 * If apply returns a non-zero status, then hashtab_map will cease
 * iterating through the hash table and will propagate the error
 * return to its caller.
 */
int hashtab_map(struct hashtab *h,
		int (*apply)(void *k, void *d, void *args),
		void *args);

/* Fill info with some hash table statistics */
void hashtab_stat(struct hashtab *h, struct hashtab_info *info);

void hashtab_cache_init(void);
#endif	/* _HASHTAB_H */

