/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: log_queue.c
* Version:  1.0
* Function Description:
*       Use list to implement a queue to hold the drop log
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/

#include <linux/list.h>
#include <linux/kthread.h>
#include <linux/netlink.h>
#include "log_queue.h"
#include "nf_hook.h"
#include "common.h"
#include "netlink.h"
#include "policy_list.h"
#include "hashtab.h"

#define MAX_QUEUE_LEN 1000
/*
 * Define a queue
 */
struct log_queue log_send_queue;
int kthread_sending_log_exit = 0;

/*
 *  Log queue initial
 *  Create a queue using a list to hold drop log information.
 */
void log_queue_init(void)
{
    INIT_LIST_HEAD(&log_send_queue.head);
    log_send_queue.len = 0;
    init_waitqueue_head(&log_send_queue.wait);
    spin_lock_init(&log_send_queue.lock);
}

/*
 *  Put the log data in the queue
 *  Call in netfilter hook functions
 *  @data: the queue element
 *  Return 0 is success
 */
int log_enqueue(struct log_queue_element *data)
{
	unsigned long flags;
    struct log_queue_element *new_data;

    if (log_send_queue.len >= MAX_QUEUE_LEN) {
        return -1;
    }

    new_data = kmalloc(sizeof(struct log_queue_element), GFP_KERNEL);
    if (!new_data) {
        return -ENOMEM;
    }

    memcpy(new_data, data, sizeof(struct log_queue_element));

    spin_lock_irqsave(&log_send_queue.lock, flags);
    list_add_tail(&new_data->list, &log_send_queue.head);
    log_send_queue.len++;
    spin_unlock_irqrestore(&log_send_queue.lock, flags);

    wake_up_interruptible(&log_send_queue.wait);

    return 0;
}

/*
 *  Get a log data from the queue
 *  Return a pointer to the queue element.
 */
static struct log_queue_element  *log_dequeue(void)
{
	unsigned long flags;
    struct log_queue_element *data = NULL;

    spin_lock_irqsave(&log_send_queue.lock, flags);
    if (!list_empty(&log_send_queue.head)) {
        data = list_first_entry(&log_send_queue.head, struct log_queue_element, list);
        list_del(&data->list);
        log_send_queue.len--;
    }
    spin_unlock_irqrestore(&log_send_queue.lock, flags);

    return data;
}
/*
 *	Destroy the queue
 */

void log_queue_destroy(void)
{
	struct log_queue_element *data = NULL;
	
	while (!log_send_queue.len) {
		data = log_dequeue();

		if (!data) {
			log_send_queue.len--;
			kfree(data);
		}
	}
	
}

/*
 * The kernel thread for sending logs to user space daemon.
 */

int drop_log_sending(void *arg)
{
    int ret = 0;
    struct log_queue_element *data;

    while (!kthread_should_stop()) {
        wait_event_interruptible(log_send_queue.wait, !list_empty(&log_send_queue.head) || kthread_sending_log_exit == 1);
        data = log_dequeue();
        if (data) {
retry:
            ret = send_usrmsg((char*)&data->log, sizeof(struct netflow_drop_log), USER_DROPLOG_PORT);
			if (ret < 0) {
				pr_err("drop log sending error(%d)\n", ret);
			} else {
				if (ret != sizeof(struct nlmsghdr) + sizeof(struct netflow_drop_log)) {
					pr_err("The size of the drop log sent is incorrect, the size is %d\n", ret);
					goto retry;
				}
			}

			kfree(data);
            data = NULL;
        }
    }

    return 0;
}

