/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: netflow.c
* Version:  1.0
* Function Description:
*       sm netflow kernel module init and exit function.
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/kthread.h>
#include <linux/workqueue.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/sched/signal.h>
#include <linux/version.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Shengye Wang");
MODULE_DESCRIPTION("sm Network flow control.");

int debug = 0;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0))
#pragma message "Kernel version is compatible."
#include "common.h"
#include "nf_hook.h"
#include "netlink.h"
#include "tuple_queue.h"
#include "log_queue.h"
#include "policy_list.h"
#include "hashtab.h"
#include "netflow_fs.h"

module_param(debug, int, S_IRUSR | S_IWUSR);
MODULE_PARM_DESC(debug, "Control the print info");

int print_drop_log = 0;
module_param(print_drop_log, int, S_IRUSR | S_IWUSR);
MODULE_PARM_DESC(debug, "Control the netflow drop log print");




#define CHECK_TASK_INTERVAL (1000*10) //10 seconds

//static struct workqueue_struct *check_task_wq;
static struct delayed_work check_task_work;
static struct timer_list check_task_timer;

static struct task_struct *kthread_handler_policy = NULL;
static struct task_struct *kthread_sending_log = NULL;

void check_task(struct work_struct *work)
{
    /* Check whether process of sm-netflow-daemon
	 * exits in the  system. if no, clear the policy
	 * list and policy hash table cache.
	*/
	struct task_struct *p;
	char buf[TASK_COMM_LEN];
	
	for_each_process(p) {
		get_task_comm(buf, p);
		if (!strncmp(buf,"sm-netflow-daem", TASK_COMM_LEN)) {
			if (debug)
				pr_info("Found the task sm-netflow-daemon.\n");
			return;
		}
	}
	
	if (debug)
		pr_info("Not found the task sm-netflow-daemon.\n");
	
	if (NO_ERROR == clear_drop_policy(&policy_list_head)) {
		hashtab_cleanup(policy_hashtab);
	}
	
	return;
}

void my_timer_callback(struct timer_list * t)
{
    /* Start the system wq task entry*/
    schedule_delayed_work(&check_task_work, 0);    
    /* reset timer */
    mod_timer(&check_task_timer, jiffies + msecs_to_jiffies(CHECK_TASK_INTERVAL));
}


int init_check_task(void)
{   
    /*Initializes the delayed work */
    INIT_DELAYED_WORK(&check_task_work, check_task);
    /*Initializes the task check timer*/
    timer_setup(&check_task_timer, my_timer_callback, 0);
    check_task_timer.expires = jiffies + msecs_to_jiffies(CHECK_TASK_INTERVAL);
    add_timer(&check_task_timer);
    
    return 0;
}

void cleanup_check_task(void)
{
    /* delete timer */
    del_timer_sync(&check_task_timer);

    /*Cancels the task from the work queue */
    cancel_delayed_work_sync(&check_task_work);
}

static int __init init_netflow(void)
{
	int ret;
	
	/*initial five-tuple queue*/
	tuple_queue_init();

	/*initial log sending queue*/
	log_queue_init();
	
	/*initial policy hash table.*/
	hashtab_cache_init();

	/*create a policy hash table*/
	policy_hashtab = hashtab_create(hash_value, key_cmp, HASHTB_SLOT_MAX);
	
	/*initial policy list*/
	init_policy_list(&policy_list_head);
	 
	/* create netlink socket */
	ret = netlink_create();

	if (ret < 0) {
		pr_err("netlink create fail, goto end.\n");
		goto end;
	}

    /*register netfilter hook functions*/
	ret = register_nf_hook(); 
	
	if (ret < 0) {
		pr_err("register nf hook fail\n");
		goto end;
	}

	init_check_task();
	//pr_info("register nf hook of sm_netlow module.\n");

	/* create the kthread of policy hash table update.*/
	kthread_handler_policy = kthread_run(updata_htable_cache, NULL, "sm_netlow_cache_update");
	if (IS_ERR(kthread_handler_policy)) {
		pr_err("sm_netlow_cache_update kthread create failed!\n");
		ret = PTR_ERR(kthread_handler_policy);
		kthread_handler_policy = NULL;
		goto end;
	}
	
	pr_info("sm_netlow_cache_update kthread create success.\n");

	/*create the kthread of drop log sending*/
	kthread_sending_log =  kthread_run(drop_log_sending, NULL, "sm_netlow_log_sending");
	if (IS_ERR(kthread_sending_log)) {
		pr_err("sm_netlow_log_sending kthread create failed!\n");
		ret = PTR_ERR(kthread_sending_log);
		kthread_sending_log = NULL;
		goto end;
	}
	netflow_fs_init();
	pr_info("sm_netlow_log_sending kthread create success.\n");

end:
	return ret;
}

static void __exit exit_netflow(void)
{
	cleanup_check_task();

	kthread_handler_policy_exit = 1;
	wake_up_interruptible(&queue.wait);
	if (kthread_handler_policy)
		kthread_stop(kthread_handler_policy);

	kthread_sending_log_exit = 1;
	wake_up_interruptible(&log_send_queue.wait);

	if (kthread_sending_log)
		kthread_stop(kthread_sending_log);

	unregister_nf_hook();
	tuple_queue_destroy();
	destroy_policy_list(&policy_list_head);
	hashtab_destroy(policy_hashtab);
	log_queue_destroy();
	netlink_release();
	netflow_fs_exit();
	pr_info("sm_netlow module exit.\n");
}

#else
#warning "Kernel version is not comptible"

static int __init init_netflow(void)
{

    printk(KERN_INFO "Netflow ktprobe register is not comptible with Current Kernel Version !\n");
    return 0;
}

static void __exit exit_netflow(void)
{

    printk(KERN_INFO "Netflow ktprobe unregister is not comptible with Current Kernel Version !\n");
}

#endif


module_init(init_netflow);
module_exit(exit_netflow);

