/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: netlink.h
* Version:  1.0
* Function Description:
*       Implement netlink network protocol creation, data sending and receiving
*       and other functions.
*
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/skbuff.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include "common.h"
#include "hashtab.h"
#include "policy_list.h"
#include "netlink.h"

#define BUF_SIZE  64
#define NETLINK_CONFIG_FILE "/etc/sm_netflow/netlink"

static int netlink_num;
static struct sock *nlsk = NULL;
extern struct net init_net;

/*
 * read netlink number form config file
 * @filename: config file
 * return netlink number str.
 */
char *read_netlink_num_form_file(char *filename)
{
	char *buf;
	ssize_t len;
	struct file *file;

	file = filp_open(filename, O_RDONLY, 0);
	if (IS_ERR(file)) {
		pr_err("sm_netflow: failed to open file %s\n", filename);
		return NULL;
	}

	buf = kmalloc(BUF_SIZE, GFP_KERNEL);
	if (!buf) {
		pr_err("sm_netflow: failed to allocate memory\n");
		filp_close(file, NULL);
		return NULL;
	}

	len = kernel_read(file, buf, BUF_SIZE, &file->f_pos);
	if (len < 0) {
		pr_err("sm_netflow: failed to read file, len = %ld\n", len);
		kfree(buf);
		filp_close(file, NULL);
		return NULL;
	}

	buf[len] = '\0';

	filp_close(file, NULL);

	return buf;
}

/*
 * send message to user space
 * @pbuf: message
 * @port: user space port
 * return size of send.
 */

int send_usrmsg(char *pbuf, uint16_t len, __u32 port)
{
    int ret;
    struct sk_buff *nl_skb;
    struct nlmsghdr *nlh;

    /*Create sk_buff using nlmsg_new().*/
    nl_skb = nlmsg_new(len, GFP_ATOMIC);
    if(!nl_skb) {
        pr_err("netlink alloc failure\n");
        return -ENOMEM;
    }

    /*Set up nlmsghdr.*/
    nlh = nlmsg_put(nl_skb, 0, 0, netlink_num, len, 0);
    if(!nlh) {
        pr_err("nlmsg_put failaure \n");
		/*If nlmsg_put() failed, nlmsg_free() will free sk_buff.*/
        nlmsg_free(nl_skb);
        return -EMSGSIZE;
    }

    /*Copy pbuf to nlmsghdr payload.*/
    memcpy(nlmsg_data(nlh), pbuf, len);
    ret = netlink_unicast(nlsk, nl_skb, port, MSG_DONTWAIT);

    return ret;
}
/*
 * recvive message from user space
 * @__skb: socket buffer
 */
static void netlink_rcv_msg(struct sk_buff *__skb)
{
	int ret = 0;
    struct sk_buff *skb = NULL;
    struct nlmsghdr *nlh = NULL;
    struct policy_recv_msg *umsg = NULL;
    struct policy_response_msg kmsg;

	skb = skb_get(__skb);

    if(skb->len >= nlmsg_total_size(0))
    {
		/* Get nlmsghdr from sk_buff. */
        nlh = nlmsg_hdr(skb);
		/* Get payload from nlmsghdr. */
        umsg = NLMSG_DATA(nlh);

		if (umsg) {
		    switch(umsg->opt) {
				case POLICY_ADD:
					ret = add_drop_policy(&policy_list_head, umsg->data);
					break;
				case POLICY_DEL:
					ret = delete_drop_policy(&policy_list_head, umsg->data);
					break;
				case POLICY_CLEAR:
					ret = clear_drop_policy(&policy_list_head);
					if (ret == NO_ERROR) 
						hashtab_cleanup(policy_hashtab);
					break;
				case POLICY_GETALL:
				{
					unsigned int policy_count = 0;
					struct netflow_drop_policy * policies = NULL;

					policies = get_all_policies(&policy_list_head, &policy_count);

					if (policies == NULL || policy_count == 0) {
						pr_info("sm_netflow: the policies list is NULL\n");
						ret = LIST_EMPTY;
					}

					kmsg.opt_return_code = ret;
					kmsg.opt = umsg->opt;

					pr_info("sm_netflow: kernel recv msg from user: opt:%d, opt_return_code:%d\n",
								umsg->opt, kmsg.opt_return_code);

					ret = send_usrmsg((char*)&kmsg, sizeof(struct policy_response_msg), USER_POLICY_PORT);
					if (ret < 0)
						pr_err("sm_netflow: kernel send msg failed.\n");

					if (!kmsg.opt_return_code && ret > 0) {
						ret = send_usrmsg((char*)policies, sizeof(struct netflow_drop_policy) * policy_count, USER_POLICY_PORT);
						if (ret < 0)
							pr_err("sm_netflow: kernel send policy data failed.\n");
					}

					if (policies)
						kfree(policies);
				}
				goto end;
				case POLICY_MAX:
					break;
			}

			kmsg.opt_return_code = ret;
			kmsg.opt = umsg->opt;

			pr_info("sm_netflow: kernel recv msg from user: opt:%d, opt_return_code:%d\n",
						umsg->opt, kmsg.opt_return_code);
            ret = send_usrmsg((char*)&kmsg, sizeof(struct policy_response_msg), USER_POLICY_PORT);
			if (ret < 0)
				pr_err("sm_netflow: kernel send msg failed.\n");
        }
    }
end:
	kfree_skb(skb);
}

struct netlink_kernel_cfg cfg = {
        .input  = netlink_rcv_msg, /* set recv callback */
};

int netlink_create(void)
{
	char *buf;
	buf = read_netlink_num_form_file(NETLINK_CONFIG_FILE);
	if (!buf) {
        pr_err("sm_netflow: read netlink config error!\n");
		return -ENOENT;
	}
	kstrtoint(buf, 0, &netlink_num);
    pr_info("sm_netflow: netlink_num = %d\n", netlink_num);
	kfree(buf);

	/* Create netlink socket */
    nlsk = (struct sock *)netlink_kernel_create(&init_net, netlink_num, &cfg);
    if(nlsk == NULL)
    {
        pr_err("sm_netflow netlink_kernel_create error !\n");
        return -ENOMEM;
    }
    pr_info("sm_netflow netlink init\n");

    return 0;
}

int netlink_release(void)
{
    if (nlsk){
        netlink_kernel_release(nlsk); /* release ..*/
        nlsk = NULL;
    }
    pr_info("sm_netflow netlink exit!\n");
	return 0;
}

