/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: nf_hook.c
* Version:  1.0
* Function Description:
*       The traffic drop logic code in the netfilter hooks.
*
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/sched.h>
#include <linux/ktime.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/version.h>
#include "common.h"
#include "nf_hook.h"
#include "hashtab.h"
#include "tuple_queue.h"
#include "log_queue.h"
#include "policy_list.h"

static unsigned int nf_ipv4_in_hook(void *priv,
		struct sk_buff *skb, const struct nf_hook_state *s);

static unsigned int nf_ipv4_out_hook(void *priv,
		struct sk_buff *skb, const struct nf_hook_state *s);

static unsigned int nf_ipv6_in_hook(void *priv,
		struct sk_buff *skb, const struct nf_hook_state *s);

static unsigned int nf_ipv6_out_hook(void *priv,
		struct sk_buff *skb, const struct nf_hook_state *s);

static struct nf_hook_ops nf_netflow_ops[] __read_mostly = {
	{
		.hook = nf_ipv4_in_hook,
		.pf = NFPROTO_IPV4,
		.hooknum = NF_INET_LOCAL_IN,
		.priority = NF_IP_PRI_FIRST,
	},
	{
		.hook = nf_ipv4_out_hook,
		.pf = NFPROTO_IPV4,
		.hooknum = NF_INET_LOCAL_OUT,
		.priority = NF_IP_PRI_FIRST,
	},
	{
		.hook = nf_ipv6_in_hook,
		.pf = NFPROTO_IPV6,
		.hooknum = NF_INET_LOCAL_IN,
		.priority = NF_IP_PRI_FIRST,
	},
	{
		.hook = nf_ipv6_out_hook,
		.pf = NFPROTO_IPV6,
		.hooknum = NF_INET_LOCAL_OUT,
		.priority = NF_IP_PRI_FIRST,
	},


};

void hdr_dump(struct ethhdr *ehdr)
{
    pr_info("[MAC_DES:%x,%x,%x,%x,%x,%x" 
           "MAC_SRC: %x,%x,%x,%x,%x,%x Prot:%x]\n",
           ehdr->h_dest[0],ehdr->h_dest[1],ehdr->h_dest[2],ehdr->h_dest[3],
           ehdr->h_dest[4],ehdr->h_dest[5],ehdr->h_source[0],ehdr->h_source[1],
           ehdr->h_source[2],ehdr->h_source[3],ehdr->h_source[4],
           ehdr->h_source[5],ehdr->h_proto);
}

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]
#define NIPQUAD_FMT "%u.%u.%u.%u"

#define NIP6(addr) \
	ntohs((addr).s6_addr16[0]), \
	ntohs((addr).s6_addr16[1]), \
	ntohs((addr).s6_addr16[2]), \
	ntohs((addr).s6_addr16[3]), \
	ntohs((addr).s6_addr16[4]), \
	ntohs((addr).s6_addr16[5]), \
	ntohs((addr).s6_addr16[6]), \
	ntohs((addr).s6_addr16[7])
#define NIP6_FMT "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x"
#define NIP6_SEQFMT "%04x%04x%04x%04x%04x%04x%04x%04x"

#if defined(__LITTLE_ENDIAN)
#define HIPQUAD(addr) \
	((unsigned char *)&addr)[3], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[0]
#elif defined(__BIG_ENDIAN)
#define HIPQUAD	NIPQUAD
#else
#error "Please fix asm/byteorder.h"
#endif /* __LITTLE_ENDIAN */

/*
 * print the five-tuple information
 *
 */
void tuple_info_print(struct pkt_addr_info *pkt)
{
	struct task_struct *tsk = current;

#define __DEBUG__
#ifdef __DEBUG__
    if (pkt->ip_version == IP_ADDR_VERSION_IPV4) {
		pr_info("direction:%d, comm:%s, ip_version:%d, protocol:%d, src_ip:'"NIPQUAD_FMT"'(%d), dest_IP:'"NIPQUAD_FMT"'(%d)",
				pkt->dir, tsk->comm, pkt->ip_version, pkt->protocol, 
				NIPQUAD(pkt->saddr.ipv4), ntohs(pkt->sport), NIPQUAD(pkt->daddr.ipv4), ntohs(pkt->dport));
	} else if (pkt->ip_version == IP_ADDR_VERSION_IPV6) {
		pr_info("direction:%d, comm:%s, ip_version:%d, protocol:%d, src_ip:'"NIP6_FMT"'(%d), dest_IP:'"NIP6_FMT"'(%d)",
				pkt->dir, tsk->comm, pkt->ip_version, pkt->protocol,
				NIP6(pkt->saddr.ipv6), ntohs(pkt->sport), NIP6(pkt->daddr.ipv6), ntohs(pkt->dport));
	}
#endif

}



#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0))
#pragma message "Kernel version is compatible."


/*
 * Parse ipv4 packet five-tuple information
 * @skb:socket buffer
 * @pkt:five-tuple information
 * @direction: The traffic direction
 * return 0 is succeed, 1 is failed.
 */
static unsigned int parse_ipv4_packet(struct sk_buff *skb,
		struct pkt_addr_info *pkt, netflow_dir_t direction)
{
	struct iphdr *iph = (struct iphdr *)(skb_network_header(skb));
		
	pkt->ip_version = IP_ADDR_VERSION_IPV4;
	pkt->dir = direction;
	pkt->saddr.ipv4.s_addr = iph->saddr;
	pkt->daddr.ipv4.s_addr = iph->daddr;

	if (iph->protocol == IPPROTO_TCP || iph->protocol == IPPROTO_UDP) {
		struct tcphdr *tcph = (struct tcphdr *)skb_transport_header(skb);
		pkt->sport = tcph->source;
		pkt->dport = tcph->dest;
		pkt->protocol = iph->protocol;
	} else {
		return 1;
	}

	return 0;
}

/*
 * Parse ipv6 packet five-tuple information
 * @skb:socket buffer
 * @pkt:five-tuple information
 * @direction: The traffic direction
 * return 0 is succeed, 1 is failed.
 */
static unsigned int parse_ipv6_packet(struct sk_buff *skb,
		struct pkt_addr_info *pkt, netflow_dir_t direction)
{
    struct ipv6hdr *ipv6h = ipv6_hdr(skb);
    
	pkt->ip_version = IP_ADDR_VERSION_IPV6;
    pkt->dir = direction;
    memcpy(&pkt->saddr.ipv6, &ipv6h->saddr, sizeof(ipv6h->saddr));
    memcpy(&pkt->daddr.ipv6, &ipv6h->daddr, sizeof(ipv6h->daddr));
    
	if (ipv6h->nexthdr == IPPROTO_TCP || ipv6h->nexthdr == IPPROTO_UDP) {
        struct tcphdr *tcph = (struct tcphdr *)skb_transport_header(skb);
		pkt->sport = tcph->source;
		pkt->dport = tcph->dest;
		pkt->protocol = ipv6h->nexthdr;
	} else {
		return 1;
	}

	return 0;

}
/* Generate traffic drop log information
 * @pkt:five-tuple information
 * @policy_id: The policy id
 * return 0 is succeed.
 */
static unsigned int generate_drop_log(struct pkt_addr_info *pkt, unsigned int *policy_id)
{
	struct netflow_drop_log log;
	struct log_queue_element log_queue_data;

	log.timestamp = get_seconds();
	log.policy_id = *policy_id;
	log.pid = current->pid;
	log.direction = pkt->dir;
	log.tuple_info.src_ip_range.start = pkt->saddr;
	log.tuple_info.src_ip_range.end = pkt->saddr;
	log.tuple_info.dst_ip_range.start = pkt->daddr;
	log.tuple_info.dst_ip_range.end = pkt->daddr;
	log.tuple_info.src_port_range.start = pkt->sport;
	log.tuple_info.src_port_range.end = pkt->sport;
	log.tuple_info.dst_port_range.start = pkt->dport;
	log.tuple_info.dst_port_range.end = pkt->dport;
	log.tuple_info.protocol = pkt->protocol;
	log.tuple_info.ip_version = pkt->ip_version;

	strncpy(log.process_name, current->comm, TASK_COMM_LEN);

    if (print_drop_log && pkt->ip_version == IP_ADDR_VERSION_IPV4) {
		pr_info("[%lu], comm:%s(%d), policy_id:%u, ip_version:%d, direction:%d, protocol:%d, src_ip:'"NIPQUAD_FMT"'(%d), dest_IP:'"NIPQUAD_FMT"'(%d)",
				log.timestamp, log.process_name, log.pid, log.policy_id,
				log.tuple_info.ip_version, log.direction, log.tuple_info.protocol, 
				NIPQUAD(log.tuple_info.src_ip_range.start.ipv4), ntohs(log.tuple_info.src_port_range.start),
				NIPQUAD(log.tuple_info.dst_ip_range.start.ipv4), ntohs(log.tuple_info.dst_port_range.start));
	} else if (print_drop_log && pkt->ip_version == IP_ADDR_VERSION_IPV6) {
		pr_info("[%lu], comm:%s(%d), policy_id:%u, ip_version:%d, direction:%d, protocol:%d, src_ip:'"NIP6_FMT"'(%d), dest_IP:'"NIP6_FMT"'(%d)",
				log.timestamp, log.process_name, log.pid, log.policy_id,
				log.tuple_info.ip_version, log.direction, log.tuple_info.protocol,
				NIP6(log.tuple_info.src_ip_range.start.ipv6), ntohs(log.tuple_info.src_port_range.start),
				NIP6(log.tuple_info.dst_ip_range.start.ipv6), ntohs(log.tuple_info.dst_port_range.start));
	}
	
	log_queue_data.log = log;
	log_enqueue(&log_queue_data);

	return 0;

}

/* Match the traffic drop policy
 * @pkt:five-tuple information
 * return 0 is match succeed, 1 is failed.
 */
static unsigned int drop_policy_match(struct pkt_addr_info *pkt)
{	
    /*查询策略缓存hash表，查询到了则阻断*/
	unsigned int  *policy_id = NULL;

	policy_id = hashtab_search(policy_hashtab, pkt);

	if (policy_id)
		generate_drop_log(pkt, policy_id);

	return  policy_id ? 0 : 1;

}

/* The ipv4 ingress netfilter hook function */

static unsigned int nf_ipv4_in_hook(void *priv, struct sk_buff *skb, 
		const struct nf_hook_state *s)
{
	int ret;
	struct tuple_data tuple;
	struct pkt_addr_info pkt;

	if (policy_list_head.len == 0) {
		return NF_ACCEPT;
	}

	ret = parse_ipv4_packet(skb, &pkt, NET_FLOW_INBOUND);

	if (ret) { 
		return NF_ACCEPT;

	}
	
	memcpy(&tuple.tuple, &pkt, sizeof(struct pkt_addr_info));
	tuple_enqueue(&tuple);
	
	return drop_policy_match(&pkt) ? NF_ACCEPT : NF_DROP;
}

/* The ipv4 egress netfilter hook function */
static unsigned int nf_ipv4_out_hook(void *priv, struct sk_buff *skb, 
		const struct nf_hook_state *s)
{
	int ret;
	struct tuple_data tuple;
	struct pkt_addr_info pkt;

	if (policy_list_head.len == 0) {
		return NF_ACCEPT;
	}

	ret = parse_ipv4_packet(skb, &pkt, NET_FLOW_OUTBOUND);

	if (ret) {
		return NF_ACCEPT;
	}

	memcpy(&tuple.tuple, &pkt, sizeof(struct pkt_addr_info));
	/* add five-tuple info to queue*/
	tuple_enqueue(&tuple);

	return drop_policy_match(&pkt) ? NF_ACCEPT : NF_DROP;
}

/* The ipv4 ingress netfilter hook function */
static unsigned int nf_ipv6_in_hook(void *priv, struct sk_buff *skb, 
		const struct nf_hook_state *s)
{
	int ret;
	struct tuple_data tuple;
	struct pkt_addr_info pkt;

	if (policy_list_head.len == 0) {
		return NF_ACCEPT;
	}

	ret = parse_ipv6_packet(skb, &pkt, NET_FLOW_INBOUND);

	if (ret) {
		return NF_ACCEPT;
	}

	memcpy(&tuple.tuple, &pkt, sizeof(struct pkt_addr_info));
	tuple_enqueue(&tuple);
	
	return drop_policy_match(&pkt) ? NF_ACCEPT : NF_DROP;
}

/* The ipv6 egress netfilter hook function */
static unsigned int nf_ipv6_out_hook(void *priv, struct sk_buff *skb, 
		const struct nf_hook_state *s)
{
	int ret;
	struct tuple_data tuple;
	struct pkt_addr_info pkt;
	
	if (policy_list_head.len == 0) {
		return NF_ACCEPT;
	}

	ret = parse_ipv6_packet(skb, &pkt, NET_FLOW_OUTBOUND);

	if (ret) {
		return NF_ACCEPT;

	}

	memcpy(&tuple.tuple, &pkt, sizeof(struct pkt_addr_info));
	tuple_enqueue(&tuple);

	return drop_policy_match(&pkt) ? NF_ACCEPT : NF_DROP;
}

/* register netfilter hook functions */
int register_nf_hook(void)
{
	int ret;
	ret = nf_register_net_hooks(&init_net, nf_netflow_ops, ARRAY_SIZE(nf_netflow_ops));
	
	if (ret < 0) {
		pr_err("sm_netflow register nf hook fail\n");
		return ret;
	}

	pr_info("sm_netflow register  nf hook\n");

	return 0;
}

/* unregister netfilter hook functions */
void unregister_nf_hook(void)
{
	nf_unregister_net_hooks(&init_net, nf_netflow_ops, ARRAY_SIZE(nf_netflow_ops));
	pr_info("sm_netflow unregister  nf hook\n");
}

#else
#warning "Kernel version is not comptible"
#endif
