/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: policy_list.c
* Version:  1.0
* Function Description:
*       Defines the implementation of related functions for linked list operations
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/
#include <linux/list.h>
#include <linux/spinlock.h>
#include <net/ipv6.h>
#include "common.h"
#include "policy_list.h"

#define POLICY_LIST_MAX  200

/*The drop policy list head*/
struct policy_list policy_list_head;

/*Initializes the list header
 *@list: the list head
 */
void init_policy_list(struct policy_list *list)
{
	INIT_LIST_HEAD(&list->head);
	spin_lock_init(&list->lock);
	list->len = 0;
}

/* Insert a new policy node before the specified head
 * @list: the list head 
 * @t: the policy entry 
 */
static inline void add_list_node(struct policy_list *list,
									struct netflow_drop_policy t)
{
    struct policy_node *new_node = kmalloc(sizeof(struct policy_node), GFP_KERNEL);
	
	spin_lock(&list->lock);
    new_node->t = t;
    list_add_tail(&(new_node->list), &list->head);
	list->len++;
	spin_unlock(&list->lock);
}

/*deletes policy node from list 
 * @list: the list head 
 * @node: the policy node 
 */
static inline void delete_list_node(struct policy_list *list,
										struct policy_node *node)
{
    spin_lock(&list->lock);
    list_del(&(node->list));
	list->len--;
	spin_unlock(&list->lock);
    kfree(node);
}

/* Modify an existing policy node in the list 
 * @list: the list head 
 * @node: the policy node 
 * @t: the policy entry 
 */
static inline void modify_list_node(struct policy_list *list,
										struct policy_node *node,
										struct netflow_drop_policy t)
{
    spin_lock(&list->lock);
    node->t = t;
	spin_unlock(&list->lock);
}

/* Finds whether the policy exits in the list. if so, returns a pointer to the policy node.
 * Otherwise, return NULL.
 * @policy_list: the list head 
 * @t: the policy entry 
 */
static inline struct policy_node *find_list_node(struct policy_list *policy_list, 
													struct netflow_drop_policy t)
{
    struct policy_node *node;
    list_for_each_entry(node, &policy_list->head, list) {
        if (node->t.policy_id == t.policy_id &&
            node->t.direction == t.direction &&
            memcmp(&node->t.tuple, &t.tuple, sizeof(t.tuple)) == 0) {
            return node;
        }
    }
    return NULL;
}

/* Finds whether the policy id exits in the list. if so, returns 0.
 * Otherwise, return -1.
 * @policy_list: the list head
 * @id: the policy id
 */
static inline int is_policy_exits(struct policy_list *policy_list,
									struct netflow_drop_policy t)
{
    struct policy_node *node;
    list_for_each_entry(node, &policy_list->head, list) {
        if (node->t.policy_id == t.policy_id ||
            (node->t.direction == t.direction &&
            memcmp(&node->t.tuple, &t.tuple, sizeof(t.tuple)) == 0)) {
            return 0;
        }
    }
    return -1;
}

/* 
 * Clear the policy list,free memory for each node
 * @policy_list: the list head 
 */
static void clear_list_nodes(struct policy_list *policy_list)
{
    struct policy_node *node, *next_node;
    list_for_each_entry_safe(node, next_node, &policy_list->head, list) {
        list_del(&(node->list));
		policy_list->len--;
        kfree(node);
    }
}

/*
 * Destroy the policy list.
 * @list: the list head 
 */
int destroy_policy_list(struct policy_list *list)
{
	if (!list) {
		return LIST_EMPTY;
	}
	
	clear_list_nodes(list);
	//kfree(list);

	return NO_ERROR;
}

/* 
 * Adds a policy to the list. 
 * @list: the list head 
 * @t: the policy entry 
 */
int add_drop_policy(struct policy_list *list, struct netflow_drop_policy t)
{
	int ret;
	
	if (!list) {
		return LIST_EMPTY;
	}

	if (list->len >= POLICY_LIST_MAX) {
		return LIST_FULL;
	}

	ret = is_policy_exits(list, t);

    if (!ret) {
        return NODE_ALREADY_EXISTS;
    }

	add_list_node(list, t);

	return NO_ERROR;
}
/* 
 * Delete a policy from the list. 
 * @policy_list: the list head 
 * @t: the policy entry 
 */
int delete_drop_policy(struct policy_list *policy_list, struct netflow_drop_policy t)
{
	struct policy_node *node = NULL;
	
	if (!policy_list) {
		return LIST_EMPTY;
	}
    
	list_for_each_entry(node, &policy_list->head, list) {
        if (node->t.policy_id == t.policy_id) {
			delete_list_node(policy_list, node);
			return NO_ERROR;
		}
    }

	return NO_FIND_NODE;

}
/* 
 * Update a policy in the list 
 * @list: the list head 
 * @t: the policy entry 
 */
int update_drop_policy(struct policy_list *list, struct netflow_drop_policy t)
{
	struct policy_node *node = NULL;
	
	if (!list) {
		return LIST_EMPTY;
	}

	node = find_list_node(list, t);

	if (!node) {
		return NO_FIND_NODE;
	}

	modify_list_node(list, node, t);

	return NO_ERROR;
}
/*
 * Gets all policies int the list
 * @list: the list head
 * @size: the policies count
 */

struct netflow_drop_policy * get_all_policies(struct policy_list *policy_list, unsigned int *size)
{
    unsigned int count = 0;
    struct policy_node *node;
    struct netflow_drop_policy *buf = kmalloc(sizeof(struct netflow_drop_policy) * policy_list->len, GFP_KERNEL);
    struct netflow_drop_policy *temp = buf;

	list_for_each_entry(node, &policy_list->head, list) {
		memcpy(temp, &node->t, sizeof(struct netflow_drop_policy));
		temp++;
		count++;
	}

	pr_info("sm_netflow: get all policies, count = %d, policy_list->len = %d.\n", count, policy_list->len);
	if (count != policy_list->len) {
		pr_err("sm_netflow: get all policies error, count = %d, policy_list->len = %d.\n", count, policy_list->len);
		kfree(buf);
		return NULL;
	}

    *size = count;
    return buf;
}


int is_ipv6_all_zero(struct in6_addr *ipv6_address)
{
    int i;
    for (i = 0; i < 16; i++) {
        if (ipv6_address->s6_addr[i] != 0) {
            return 0;
        }
    }
    return 1;
}


/* 
 * Clear all policy in the list. 
 * @list: the list head 
 */
int clear_drop_policy(struct policy_list *list)
{
	if (!list) {
		return LIST_EMPTY;
	}

	clear_list_nodes(list);

	return NO_ERROR;
}

/* 
 * Generate a policy to exchange source and destination addresses.
 * return a new policy pointer.
 * @t: the policy need to switched
 */
static struct netflow_drop_policy* reverse_five_tuple_policy(struct netflow_drop_policy* t)
{
	struct netflow_drop_policy* reverse_policy = kmalloc(sizeof(struct netflow_drop_policy), GFP_KERNEL);
	if (!reverse_policy)
		return NULL;
	
	reverse_policy->policy_id = t->policy_id;
	reverse_policy->direction =  t->direction;
	reverse_policy->tuple.protocol = t->tuple.protocol;
	reverse_policy->tuple.ip_version = t->tuple.ip_version;
	reverse_policy->tuple.src_port_range.start = t->tuple.dst_port_range.start;
	reverse_policy->tuple.src_port_range.end = t->tuple.dst_port_range.end;
	reverse_policy->tuple.dst_port_range.start = t->tuple.src_port_range.start;
	reverse_policy->tuple.dst_port_range.end = t->tuple.src_port_range.end;

	if (t->tuple.ip_version == IP_ADDR_VERSION_IPV4) {
	    memcpy(&reverse_policy->tuple.src_ip_range.start.ipv4, &t->tuple.dst_ip_range.start.ipv4, sizeof(struct in_addr));
	    memcpy(&reverse_policy->tuple.src_ip_range.end.ipv4, &t->tuple.dst_ip_range.end.ipv4, sizeof(struct in_addr));
	    memcpy(&reverse_policy->tuple.dst_ip_range.start.ipv4, &t->tuple.src_ip_range.start.ipv4, sizeof(struct in_addr));
	    memcpy(&reverse_policy->tuple.dst_ip_range.end.ipv4, &t->tuple.src_ip_range.end.ipv4, sizeof(struct in_addr));
	} else if (t->tuple.ip_version == IP_ADDR_VERSION_IPV6) {
	    memcpy(&reverse_policy->tuple.src_ip_range.start.ipv6, &t->tuple.dst_ip_range.start.ipv6, sizeof(struct in6_addr));
	    memcpy(&reverse_policy->tuple.src_ip_range.end.ipv6, &t->tuple.dst_ip_range.end.ipv6, sizeof(struct in6_addr));
	    memcpy(&reverse_policy->tuple.dst_ip_range.start.ipv6, &t->tuple.src_ip_range.start.ipv6, sizeof(struct in6_addr));
	    memcpy(&reverse_policy->tuple.dst_ip_range.end.ipv6, &t->tuple.src_ip_range.end.ipv6, sizeof(struct in6_addr));
	} else {
		kfree(reverse_policy);
		return NULL;
	}

	return reverse_policy;
}

/*
 * If the five-tuple int the packet matches the policy five-tuple,
 * the ID of the policy is returned successfully. Otherwise, 0 is returned.
 */
static unsigned int compare_five_tuple(struct netflow_drop_policy* t,
										struct pkt_addr_info pkt)
{
	if (t->tuple.ip_version == pkt.ip_version && t->tuple.protocol == pkt.protocol &&
			((memcmp(&pkt.sport, &t->tuple.src_port_range.start, sizeof(pkt.sport)) >= 0 && 
			memcmp(&pkt.sport, &t->tuple.src_port_range.end, sizeof(pkt.sport)) <= 0) ||
			(t->tuple.src_port_range.start == 0 && t->tuple.src_port_range.end == 0)) &&
			((memcmp(&pkt.dport, &t->tuple.dst_port_range.start, sizeof(pkt.dport)) >= 0 &&
			memcmp(&pkt.dport, &t->tuple.dst_port_range.end, sizeof(pkt.dport)) <= 0) ||
			(t->tuple.dst_port_range.start == 0 && t->tuple.dst_port_range.end == 0))) 
	{
		if (pkt.ip_version == IP_ADDR_VERSION_IPV4) {
			if (((t->tuple.src_ip_range.start.ipv4.s_addr == 0 &&
				t->tuple.src_ip_range.end.ipv4.s_addr == 0) ||
				(memcmp(&pkt.saddr.ipv4, &t->tuple.src_ip_range.start.ipv4, sizeof(struct in_addr)) >= 0 &&
				memcmp(&pkt.saddr.ipv4, &t->tuple.src_ip_range.end.ipv4, sizeof(struct in_addr)) <= 0)) &&
				((t->tuple.dst_ip_range.start.ipv4.s_addr == 0 &&
				t->tuple.dst_ip_range.end.ipv4.s_addr == 0) ||
				(memcmp(&pkt.daddr.ipv4, &t->tuple.dst_ip_range.start.ipv4, sizeof(struct in_addr)) >= 0 &&
				memcmp(&pkt.daddr.ipv4, &t->tuple.dst_ip_range.end.ipv4, sizeof(struct in_addr)) <= 0)))
					
				return t->policy_id;
			
		} else if (pkt.ip_version == IP_ADDR_VERSION_IPV6) {
			if (((ipv6_addr_any(&t->tuple.src_ip_range.start.ipv6) &&
				ipv6_addr_any(&t->tuple.src_ip_range.end.ipv6)) ||
				(memcmp(&pkt.saddr.ipv6, &t->tuple.src_ip_range.start.ipv6, sizeof(struct in6_addr)) >= 0 &&
				memcmp(&pkt.saddr.ipv6, &t->tuple.src_ip_range.end.ipv6, sizeof(struct in6_addr)) <= 0)) &&
				((ipv6_addr_any(&t->tuple.dst_ip_range.start.ipv6) &&
				ipv6_addr_any(&t->tuple.dst_ip_range.end.ipv6)) ||
				(memcmp(&pkt.daddr.ipv6, &t->tuple.dst_ip_range.start.ipv6, sizeof(struct in6_addr)) >= 0 &&
				memcmp(&pkt.daddr.ipv6, &t->tuple.dst_ip_range.end.ipv6, sizeof(struct in6_addr)) <= 0)))

				return t->policy_id;
		}
	}

	return 0;	
}

/* Matches the policy by traversing the list, if found, return the policy id,
 * Otherwise return -1.
 * @list: the list head 
 * @pkt: the information such as IP addresses obtained from network traffic
 */
unsigned int match_drop_polciy(struct policy_list *policy_list,
								struct pkt_addr_info pkt) {
	
    unsigned int ret = 0;
	struct policy_node *node;
	struct netflow_drop_policy* new_policy = NULL;

	list_for_each_entry(node, &policy_list->head, list) {
		if (node->t.direction == pkt.dir) {
			ret = compare_five_tuple(&node->t, pkt);
			if (ret > 0)
				break;
		} else if (node->t.direction == NET_FLOW_ALL) {
			ret = compare_five_tuple(&node->t, pkt);
			if (ret > 0) {
				break;
			} else {
				new_policy = reverse_five_tuple_policy(&node->t);
				if (!new_policy) {
					pr_err("kmalloc error in reverse five-tuple policy.");
					break;
				}
				ret = compare_five_tuple(new_policy, pkt);
				kfree(new_policy);
				new_policy = NULL;
				if (ret > 0) {
					break;
				}
			}
		}
	}

	return ret;
}

