/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: tuple_queue.c
* Version:  1.0
* Function Description:
*       Use list to implement a queue to hold the five-tuple 
*		information parsed from network traffic
* History:
*    <author>     <time>      <version>    <desc>
*    wangshengye  2023/5/19    1.0         Initial build
********************************************************************************/
#include <linux/list.h>
#include <linux/kthread.h>
#include <linux/slab.h>
#include "tuple_queue.h"
#include "nf_hook.h"
#include "common.h"
#include "policy_list.h"
#include "hashtab.h"

#define MAX_QUEUE_LEN 1000

/*
 * Define a queue
 */
struct tuple_queue queue;
int kthread_handler_policy_exit = 0;
/*
 *  Five-tuple queue initial
 *  Create a queue using a list to hold five-tuple information.
 */

void tuple_queue_init(void)
{
    INIT_LIST_HEAD(&queue.head);
    queue.len = 0;
    init_waitqueue_head(&queue.wait);
    spin_lock_init(&queue.lock);
}

/*
 *	Put the five-tuple data in the queue
 *	@data: the queue element
 *  Return 0 is success
 */
int tuple_enqueue(struct tuple_data *data)
{
	unsigned long flags;
    struct tuple_data *new_data;

    if (queue.len >= MAX_QUEUE_LEN) {
        return -1;
    }

    new_data = kmalloc(sizeof(struct tuple_data), GFP_KERNEL);
    if (!new_data) {
        return -ENOMEM;
    }

    memcpy(new_data, data, sizeof(struct tuple_data));

    spin_lock_irqsave(&queue.lock, flags);
    list_add_tail(&new_data->list, &queue.head);
    queue.len++;
    spin_unlock_irqrestore(&queue.lock, flags);
    /*The waiting process is woken up when data is added to the queue.*/
    wake_up_interruptible(&queue.wait);

    return 0;
}

/*
 *  Get a five-tuple data from the queue
 *  Return a pointer to the queue element.
 */

static struct tuple_data *tuple_dequeue(void)
{
	unsigned long flags;
    struct tuple_data *data = NULL;

    spin_lock_irqsave(&queue.lock, flags);
    if (!list_empty(&queue.head)) {
        data = list_first_entry(&queue.head, struct tuple_data, list);
        list_del(&data->list);
        queue.len--;
    }
    spin_unlock_irqrestore(&queue.lock, flags);

    return data;
}
/*
 * Destroy the queue
 */

void tuple_queue_destroy(void)
{
	struct tuple_data *data = NULL;
	
	while (!queue.len) {
		data = tuple_dequeue();

		if (!data) {
			queue.len--;
			kfree(data);
		}
	}
	
}

/*
 *  Update the policy cache hash table.
 *	1.Create a separete thread to execute this function
 *  2.Get a five-tuple information for tuple queue.
 *  3.If the queue is empty, the thread enters interruptible sleep.
 *  4.The five-tuple information obtained from the tuple queue is 
 *    used for matching in the policy list. if the matching succeeds,
 *    the policy is inserted into the Hash table. Otherwise, check whether
 *    the policy exists in the hash table and delete it.
 */

int updata_htable_cache(void *arg)
{
	int ret = 0;
	unsigned int policy_id;
    struct tuple_data *data;

    while (!kthread_should_stop()) {
		wait_event_interruptible(queue.wait, !list_empty(&queue.head) || kthread_handler_policy_exit == 1);
        data = tuple_dequeue();
        if (data) {	
			policy_id = match_drop_polciy(&policy_list_head, data->tuple);
			
			if (policy_id) {
				ret = hashtab_insert(policy_hashtab, &data->tuple, sizeof(struct pkt_addr_info),  &policy_id, sizeof(unsigned int));
				if (ret == 0) { 
					if (debug) {
					    printk("insert_node success!\n");
				        tuple_info_print(&data->tuple);
					}
				} else {
					if (debug) {
						printk("insert_node failed!, ret = %d\n", ret);
						tuple_info_print(&data->tuple);
					}
				}

			} else {
				ret = hashtab_remove(policy_hashtab, &data->tuple);
				if (!ret) {
					if (debug) {
						printk("delete_node success!\n");
						tuple_info_print(&data->tuple);
					}
				}
			}

            kfree(data);
			data = NULL;
        }
    }

    return 0;
}

