/******************************************************************************
* Copyright Kylin Software  Co, Ltd.(c) (2020)
* FileName: netflow_payload.h
* Version:  1.0
* Function Description:
* A hook for get netflow payload 
*
* History:
*   <author>     <time>     <version>    <desc>
*   maguicai  2023/05/19   1.0         Initial build
********************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/cdev.h>
#include <linux/udp.h>
#include <linux/kfifo.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <linux/version.h>


#define TASK_COMM_LEN 16
#define MAX_PAYLOAD_BYTES 256
#define FIFO_SIZE 2048
#define RING_BUFFER_SIZE 2048
#define MAX_IP_HEADER 40
#define MAX_PACKET_NODES 100
enum direction { INGRESS, EGRESS };

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5,15,0))
#pragma message "Kernel version is compatible."


struct packet_data {
    pid_t pid;
    char comm[TASK_COMM_LEN];
    u8 dir; // INGRESS or EGRESS
    u8 proto; // IPPROTO_TCP or IPPROTO_UDP
    unsigned char ip_version;
    int payload_len;
    unsigned char ip_header[MAX_IP_HEADER];
    unsigned char tcpudp_header[sizeof(struct tcphdr)]; // Using the larger struct size for both TCP and UDP
    unsigned char payload[MAX_PAYLOAD_BYTES];
};
struct packet_data_list {
    struct packet_data *packet;
    struct work_struct save_work;
    int print_on ;  // for trace 
};

struct ring_buffer {
    struct packet_data buffer[RING_BUFFER_SIZE];
    unsigned int read_pos;
    unsigned int write_pos;
    struct mutex lock;
};
static struct ring_buffer rb;
static struct workqueue_struct *workqueue;
//static LIST_HEAD(packet_node_list);  //初始化队列


static struct kmem_cache *packet_data_list_cache = NULL;
static struct kmem_cache *packet_cache = NULL;

void process_netflow_payload(struct sk_buff *skb, int hook_num);
static void save_packet(struct work_struct *work);
static int major_number;
static struct class *netflow_class;

/***
 * count 代表读取多少条 packet_data数据
 * 返回值：读取了多少条数据
*/
static ssize_t netflow_chardev_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
    //printk(KERN_INFO " netflow read from ko  count=%d...\n",count);
    ssize_t total = 0 ;
    int available ;
    if (mutex_lock_interruptible(&rb.lock)) {
    return -ERESTARTSYS;
    }
    //需要检查 这个count是否有效，避免访问非法内存
    available = (rb.write_pos >= rb.read_pos) ? 
    (rb.write_pos - rb.read_pos) : (RING_BUFFER_SIZE - rb.read_pos + rb.write_pos);

    
    // 如果请求的数目比可用的数据多时，取可用数
    if (count > available) {
        count = available;

        /***
        mutex_unlock(&rb.lock);
        return -EINVAL;
        **/
    }
    //printk(KERN_INFO " netflow read from ko  count=%d...\n",count);
    


    while (count > 0 && rb.read_pos != rb.write_pos) {
        if(copy_to_user(buf + total * sizeof(struct packet_data), rb.buffer + rb.read_pos, sizeof(struct packet_data))){
		return -EFAULT;
	}
        rb.read_pos = (rb.read_pos + 1) % RING_BUFFER_SIZE;
        total++;
        count--;
    }
    mutex_unlock(&rb.lock);
    return total;
}
static const struct file_operations netflow_fops = {
    .owner = THIS_MODULE,
    .read = netflow_chardev_read,
};



static int __init chardev_init(void)
{
    major_number = register_chrdev(0, "netflow_chardev", &netflow_fops);
    if (major_number < 0)
    {
        printk(KERN_ERR "register_chrdev failed: %d\n", major_number);
        return major_number;
    }

    netflow_class = class_create(THIS_MODULE, "netflow_chardev_class");
    if (IS_ERR(netflow_class))
    {
        unregister_chrdev(major_number, "netflow_chardev");
        return PTR_ERR(netflow_class);
    }

    device_create(netflow_class, NULL, MKDEV(major_number, 0), NULL, "netflow_chardev");
    printk(KERN_INFO "Kprobe CharDEV created ...\n");

    return 0;
}

static void __exit chardev_exit(void)
{
    device_destroy(netflow_class, MKDEV(major_number, 0));
    class_destroy(netflow_class);
    unregister_chrdev(major_number, "netflow_chardev");
}

static unsigned int nf_hook_local_in(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    //printk(KERN_INFO " netflow netfilter local_in hook PID=%d COMM=%s \n", task_pid_nr(current),current->comm);
    process_netflow_payload(skb, state->hook);
    return NF_ACCEPT;
}

static unsigned int nf_hook_local_in6(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    process_netflow_payload(skb, state->hook);
    return NF_ACCEPT;
}

static unsigned int nf_hook_local_out(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    process_netflow_payload(skb, state->hook);
    return NF_ACCEPT;
}

static unsigned int nf_hook_local_out6(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    process_netflow_payload(skb, state->hook);
    return NF_ACCEPT;
}

void process_netflow_payload(struct sk_buff *skb,  int hook_num)
{
    //printk(KERN_INFO " netflow process payload start: \n");
  
   struct iphdr *iph;
    struct ipv6hdr *ipv6h;
    struct tcphdr *tcph;
    struct udphdr *udph;
    __be16 dest_port,src_port;
    unsigned char *tcpudp_payload = 0;
    int tcpudp_payload_len = 0;
    struct packet_data_list *packet_list_node ;

    struct packet_data *packet = kmem_cache_alloc(packet_cache, GFP_ATOMIC);
    if (!packet) {
        return ; // 内存分配失败，直接返回
    }else{
        memset(packet, 0, sizeof(struct packet_data));
    }
    packet_list_node = kmem_cache_alloc(packet_data_list_cache, GFP_ATOMIC);
    if (!packet_list_node) {
        printk(KERN_ERR " packet_list_node is null !!\n");
        return ;
    }else{
        memset(packet_list_node, 0, sizeof(struct packet_data_list));
    }
    packet_list_node->packet = packet;
    //packet_list_node->print_on = 0;
    INIT_WORK(&packet_list_node->save_work, save_packet);

    if (!skb || !skb->data){
        return;
    }
    iph = ip_hdr(skb);
    if (!iph){
        return;
    }
    
    if (iph->version == 4) {
        // IPv4 packet
        if (iph->protocol == IPPROTO_TCP || iph->protocol == IPPROTO_UDP) {
            packet->proto = iph->protocol;
            memcpy(packet->ip_header, iph, iph->ihl * 4);
            if (iph->protocol == IPPROTO_TCP) {
                //printk(KERN_INFO " Kprobe Protocol=%d\n",iph->protocol);
                tcph = tcp_hdr(skb);
                if (!tcph){
                    return;
                }
                memcpy(packet->tcpudp_header, tcph, tcph->doff * 4);
                tcpudp_payload = (unsigned char *)((unsigned char *)tcph + (tcph->doff * 4));
                tcpudp_payload_len = ntohs(iph->tot_len) - (iph->ihl * 4) - (tcph->doff * 4);
                // for test print 
                dest_port = tcph->dest;
                src_port = tcph->source;
            } else {
                udph = udp_hdr(skb);
                memcpy(packet->tcpudp_header, udph, sizeof(struct udphdr));
                tcpudp_payload = (unsigned char *)((unsigned char *)udph + sizeof(struct udphdr));
                tcpudp_payload_len = ntohs(iph->tot_len) - (iph->ihl * 4) - sizeof(struct udphdr);
            }
            packet->ip_version = 4;
        }
    } else if (iph->version == 6) {
        // IPv6 packet
        ipv6h = ipv6_hdr(skb);
        if (ipv6h->nexthdr == IPPROTO_TCP || ipv6h->nexthdr == IPPROTO_UDP) {
            packet->proto = ipv6h->nexthdr;
            memcpy(packet->ip_header, ipv6h, sizeof(struct ipv6hdr));
            if (ipv6h->nexthdr == IPPROTO_TCP) {
                //printk(KERN_INFO " Kprobe Protocol=%d\n",ipv6h->nexthdr);
                tcph = tcp_hdr(skb);
                memcpy(packet->tcpudp_header, tcph, tcph->doff * 4);
                tcpudp_payload = (unsigned char *)((unsigned char *)tcph + (tcph->doff * 4));
                tcpudp_payload_len = ntohs(ipv6h->payload_len) - (tcph->doff * 4);
            } else {
                //printk(KERN_INFO " Kprobe Protocol=%d\n",ipv6h->nexthdr);
                udph = udp_hdr(skb);
                memcpy(packet->tcpudp_header, udph, sizeof(struct udphdr));
                tcpudp_payload = (unsigned char *)((unsigned char *)udph + sizeof(struct udphdr));
                tcpudp_payload_len = ntohs(ipv6h->payload_len) - sizeof(struct udphdr);
            }
        packet->ip_version = 6;
    }
    } else {
        return;
    }

    if (tcpudp_payload_len > MAX_PAYLOAD_BYTES) {
        tcpudp_payload_len = MAX_PAYLOAD_BYTES;
    }
    packet->payload_len = tcpudp_payload_len ;
    if (tcpudp_payload_len > 0){
        memcpy(packet->payload, tcpudp_payload, tcpudp_payload_len);
    }
    // Capture the process information
    packet->pid = task_pid_nr(current);
    strncpy(packet->comm, current->comm, TASK_COMM_LEN);
    packet->comm[TASK_COMM_LEN - 1] = '\0';
    // Determine packet direction
    packet->dir= hook_num;

    
    queue_work(workqueue, &packet_list_node->save_work);
    //printk(KERN_INFO " netflow payload write_pos=%d read_pos=%d \n",rb.write_pos,rb.read_pos);
}

static void save_packet(struct work_struct *work) {
    
     struct packet_data_list *packet_list_node ;
     struct packet_data *packet ;
     packet_list_node = container_of(work, struct packet_data_list, save_work);
     if(packet_list_node == NULL)
     {
        printk(KERN_ERR " save_packet error packet_list_node is NULL !! \n");
        return;
     }
     packet = packet_list_node->packet;
     if(packet == NULL)
     {
        printk(KERN_ERR " save_packet error packet_list_node->packet is NULL !! \n");
        return ;
     }

    // 在这里执行数据保存的操作，
    mutex_lock(&rb.lock);
    //rb.buffer[rb.write_pos] = *packet;
    memcpy(&rb.buffer[rb.write_pos],packet_list_node->packet,sizeof(struct packet_data));
    rb.write_pos = (rb.write_pos + 1) % RING_BUFFER_SIZE;
    //检查 是否覆盖了读指针
    if(rb.write_pos == rb.read_pos){
        rb.read_pos = (rb.read_pos + 1) % RING_BUFFER_SIZE;
    }
    kmem_cache_free(packet_data_list_cache, packet_list_node);
    kmem_cache_free(packet_cache, packet);
    //kfree(packet);
    //kfree(packet_list);
    mutex_unlock(&rb.lock);
    //printk(KERN_INFO " Saving packet S6 : save_packet end  \n");
}

static struct nf_hook_ops nfho_local_in = {
    .hook = nf_hook_local_in,
    .hooknum = NF_INET_LOCAL_IN,
    .pf = PF_INET,
    .priority = NF_IP_PRI_FIRST,
};
static struct nf_hook_ops nfho_local_in6 = {
    .hook = nf_hook_local_in6,
    .hooknum = NF_INET_LOCAL_IN,
    .pf = PF_INET6,
    .priority = NF_IP_PRI_FIRST,
};

static struct nf_hook_ops nfho_local_out = {
    .hook = nf_hook_local_out,
    .hooknum = NF_INET_LOCAL_OUT,
    .pf = PF_INET,
    .priority = NF_IP_PRI_FIRST,
};
static struct nf_hook_ops nfho_local_out6 = {
    .hook = nf_hook_local_out6,
    .hooknum = NF_INET_LOCAL_OUT,
    .pf = PF_INET6,
    .priority = NF_IP_PRI_FIRST,
};

static int __init nf_hooks_init(void)
{
    int ret;

    // 初始化ring_buffer 
    memset(rb.buffer, 0, sizeof(rb.buffer));
    rb.read_pos = 0;
    rb.write_pos = 0;
    mutex_init(&rb.lock);


    // for workqueue
    // 创建异步工作队列
    workqueue = create_singlethread_workqueue("save_workqueue");
    if (!workqueue) {
        return -ENOMEM;
    }

    // 初始化保存数据的工作


    // 初始 化 kmem_cache 
    packet_cache = kmem_cache_create("packet_cache", sizeof(struct packet_data), 0, SLAB_HWCACHE_ALIGN, NULL);
    if (!packet_cache)
        return -ENOMEM;
    packet_data_list_cache = kmem_cache_create("packet_data_cache",
                                          sizeof(struct packet_data),
                                          0,
                                          SLAB_HWCACHE_ALIGN, 
                                          NULL);
    if (!packet_data_list_cache)
        return -ENOMEM;


    ret = nf_register_net_hook(&init_net, &nfho_local_in);
    if (ret < 0)
    {
        printk(KERN_INFO "nf_register_net_hook (local_in) failed, returned %d\n", ret);
        return ret;
    }
    ret = nf_register_net_hook(&init_net, &nfho_local_in6);
    if (ret < 0)
    {
        printk(KERN_INFO "nf_register_net_hook (local_in) failed, returned %d\n", ret);
        return ret;
    }

    ret = nf_register_net_hook(&init_net, &nfho_local_out);
    if (ret < 0)
    {
        printk(KERN_INFO "nf_register_net_hook (local_out) failed, returned %d\n", ret);
        nf_unregister_net_hook(&init_net, &nfho_local_in);
        return ret;
    }
    ret = nf_register_net_hook(&init_net, &nfho_local_out6);
    if (ret < 0)
    {
        printk(KERN_INFO "nf_register_net_hook (local_out) failed, returned %d\n", ret);
        nf_unregister_net_hook(&init_net, &nfho_local_in6);
        return ret;
    }

    printk(KERN_INFO "Kprobe Netfilter hooks registered\n");
    ret = chardev_init();
    printk(KERN_INFO "Kprobe Netfilter CharDev init..\n");
    return 0;
}

static void __exit nf_hooks_exit(void)
{
    nf_unregister_net_hook(&init_net, &nfho_local_in);
    nf_unregister_net_hook(&init_net, &nfho_local_out);
    nf_unregister_net_hook(&init_net, &nfho_local_in6);
    nf_unregister_net_hook(&init_net, &nfho_local_out6);


    // 清理保存数据的工作
    //cancel_delayed_work_sync(&save_work);

    // 销毁异步工作队列
    if (workqueue) {
        flush_workqueue(workqueue);
        destroy_workqueue(workqueue);
    }

    // 释放队列中的数据包
    /**  不需要使用队列了
    struct packet_data_list *packet_node, *tmp;
    list_for_each_entry_safe(packet_node, tmp, &packet_node_list, list) {
        list_del(&packet_node->list);
        kfree(packet_node);
    }
    **/
   if (packet_cache)
        kmem_cache_destroy(packet_cache);
    if (packet_data_list_cache)
        kmem_cache_destroy(packet_data_list_cache);

    // 清理对象池
    /**
    int i;
    for (i = 0; i < packet_node_pool_index; i++) {
        // 可以进行一些额外的清理操作，如取消工作队列等
        
        // 清空 packet_node 的状态
        //packet_node_pool[i].packet = NULL;
        cancel_work_sync(&packet_node_pool[i].save_work);
        //INIT_LIST_HEAD(&packet_node_pool[i].list);
        free_packet_node(&packet_node_pool[i]);
    }
    **/

    printk(KERN_INFO "Kprobe Netfilter hooks unregistered\n");
    chardev_exit();
    printk(KERN_INFO "Kprobe Netfilter CharDEV unregistered\n");
}

#else
#warning "Kernel version is not comptible"

static int __init nf_hooks_init(void)
{

    printk(KERN_INFO "Netflow ktprobe register is not comptible with Current Kernel Version !\n");
    return 0;
}

static void __exit nf_hooks_exit(void)
{

    printk(KERN_INFO "Netflow ktprobe unregister is not comptible with Current Kernel Version !\n");
}

#endif



module_init(nf_hooks_init);
module_exit(nf_hooks_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kylin");
MODULE_DESCRIPTION("Netfilter hooks for Netflow payload");
